"""Classification Estimator"""
# pylint: disable=unused-variable,bad-whitespace,missing-function-docstring,logging-format-interpolation,arguments-differ,logging-not-lazy
import time
import os
import math
import copy

from PIL import Image
import pandas as pd
import numpy as np
import mxnet as mx
from mxnet import gluon, nd
from mxnet.optimizer import Optimizer
from mxnet import autograd as ag
from mxnet.gluon.data.vision import transforms
from ....data.transforms.presets.imagenet import transform_eval
from ....model_zoo import get_model
from ....utils import LRSequential, LRScheduler
from .... import nn
from .... import loss
from ..base_estimator import BaseEstimator, set_default
from ..constants import BINARY, MULTICLASS, REGRESSION
from .utils import get_data_loader, get_data_rec, smooth, mixup_transform
from .default import ImageClassificationCfg
from ...data.dataset import ImageClassificationDataset
from ..conf import _BEST_CHECKPOINT_FILE
from ..utils import EarlyStopperOnPlateau

__all__ = ['ImageClassificationEstimator']


@set_default(ImageClassificationCfg())
class ImageClassificationEstimator(BaseEstimator):
    """Estimator implementation for Image Classification.

    Parameters
    ----------
    config : dict
        Config in nested dict.
    logger : logging.Logger
        Optional logger for this estimator, can be `None` when default setting is used.
    reporter : callable
        The reporter for metric checkpointing.
    net : mx.gluon.Block
        The custom network. If defined, the model name in config will be ignored so your
        custom network will be used for training rather than pulling it from model zoo.
    """
    Dataset = ImageClassificationDataset
    def __init__(self, config, logger=None, reporter=None, net=None, optimizer=None, problem_type=None):
        super(ImageClassificationEstimator, self).__init__(config, logger=logger, reporter=reporter, name=None)
        if problem_type is None:
            problem_type = MULTICLASS
        self._problem_type = problem_type
        self.last_train = None
        self.input_size = self._cfg.train.input_size
        self._feature_net = None
        if net is not None:
            assert isinstance(net, gluon.Block), f"given custom network {type(net)}, `gluon.Block` expected"
            try:
                # to avoid cuda initialization error, we keep network copies in cpu
                net.collect_params().reset_ctx(mx.cpu())
            except ValueError:
                pass
        self._custom_net = net
        if optimizer is not None:
            if isinstance(optimizer, str):
                pass
            else:
                assert isinstance(optimizer, Optimizer)
        self._optimizer = optimizer

    def _fit(self, train_data, val_data, time_limit=math.inf):
        tic = time.time()
        self._best_acc = -float('inf')
        self.epoch = 0
        self._time_elapsed = 0
        if max(self._cfg.train.start_epoch, self.epoch) >= self._cfg.train.epochs:
            return {'time', self._time_elapsed}
        if not isinstance(train_data, pd.DataFrame):
            self.last_train = len(train_data)
        else:
            self.last_train = train_data
        self._init_trainer()
        self._time_elapsed += time.time() - tic
        return self._resume_fit(train_data, val_data, time_limit=time_limit)

    def _resume_fit(self, train_data, val_data, time_limit=math.inf):
        tic = time.time()
        if max(self._cfg.train.start_epoch, self.epoch) >= self._cfg.train.epochs:
            return {'time', self._time_elapsed}
        if self._problem_type != REGRESSION and (not self.classes or not self.num_class):
            raise ValueError('This is a classification problem and we are not able to determine classes of dataset')

        num_workers = self._cfg.train.num_workers
        if self._cfg.train.use_rec:
            self._logger.info(f'Loading data from rec files: {self._cfg.train.rec_train}/{self._cfg.train.rec_val}')
            train_loader, val_loader, self.batch_fn = get_data_rec(self._cfg.train.rec_train,
                                                                   self._cfg.train.rec_train_idx,
                                                                   self._cfg.train.rec_val,
                                                                   self._cfg.train.rec_val_idx,
                                                                   self.batch_size, num_workers,
                                                                   self.input_size,
                                                                   self._cfg.train.crop_ratio)
        else:
            train_dataset = train_data.to_mxnet()
            val_dataset = val_data.to_mxnet()
            train_loader, val_loader, self.batch_fn = get_data_loader(self._cfg.train.data_dir,
                                                                      self.batch_size, num_workers,
                                                                      self.input_size,
                                                                      self._cfg.train.crop_ratio,
                                                                      train_dataset=train_dataset,
                                                                      val_dataset=val_dataset)
        self._time_elapsed += time.time() - tic
        return self._train_loop(train_loader, val_loader, time_limit=time_limit)

    def _train_loop(self, train_data, val_data, time_limit=math.inf):
        start_tic = time.time()
        if self._cfg.train.no_wd:
            for k, v in self.net.collect_params('.*beta|.*gamma|.*bias').items():
                v.wd_mult = 0.0
        if self._cfg.train.label_smoothing or self._cfg.train.mixup:
            sparse_label_loss = False
        else:
            sparse_label_loss = True
        if self.distillation:
            L = loss.DistillationSoftmaxCrossEntropyLoss(temperature=self._cfg.train.temperature,
                                                         hard_weight=self._cfg.train.hard_weight,
                                                         sparse_label=sparse_label_loss)
        else:
            L = gluon.loss.SoftmaxCrossEntropyLoss(sparse_label=sparse_label_loss)

        if self._cfg.train.mixup:
            train_metric = mx.metric.RMSE()
        else:
            train_metric = mx.metric.Accuracy()

        if self._problem_type == REGRESSION:
            train_metric = mx.metric.RMSE()
            L = gluon.loss.L2Loss()

        if self._cfg.train.mode == 'hybrid':
            self.net.hybridize(static_alloc=True, static_shape=True)
            if self.distillation:
                self.teacher.hybridize(static_alloc=True, static_shape=True)

        self._logger.info('Start training from [Epoch %d]', max(self._cfg.train.start_epoch, self.epoch))
        early_stopper = EarlyStopperOnPlateau(
            patience=self._cfg.train.early_stop_patience,
            min_delta=self._cfg.train.early_stop_min_delta,
            baseline_value=self._cfg.train.early_stop_baseline,
            max_value=self._cfg.train.early_stop_max_value)
        train_metric_name, train_metric_score = train_metric.get()
        train_metric_score = -float('inf')
        cp_name = ''
        self._time_elapsed += time.time() - start_tic
        for self.epoch in range(max(self._cfg.train.start_epoch, self.epoch), self._cfg.train.epochs):
            epoch = self.epoch
            if self._best_acc >= 1.0:
                self._logger.info('[Epoch {}] Early stopping as acc is reaching 1.0'.format(epoch))
                break
            should_stop, stop_message = early_stopper.get_early_stop_advice()
            if should_stop:
                self._logger.info('[Epoch {}] '.format(epoch) + stop_message)
                break
            tic = time.time()
            last_tic = time.time()
            mx.nd.waitall()
            if self._cfg.train.use_rec:
                train_data.reset()
            train_metric.reset()

            # pylint: disable=undefined-loop-variable
            for i, batch in enumerate(train_data):
                btic = time.time()
                if self._time_elapsed > time_limit:
                    self._logger.warning(f'`time_limit={time_limit}` reached, exit early...')
                    return {'train_acc': train_metric_score, 'valid_acc': self._best_acc,
                            'time': self._time_elapsed, 'checkpoint': cp_name}
                data, label = self.batch_fn(batch, self.ctx)

                if self._cfg.train.mixup:
                    lam = np.random.beta(self._cfg.train.mixup_alpha,
                                         self._cfg.train.mixup_alpha)
                    if epoch >= self._cfg.train.epochs - self._cfg.train.mixup_off_epoch:
                        lam = 1
                    data = [lam*X + (1-lam)*X[::-1] for X in data]

                    if self._cfg.train.label_smoothing:
                        eta = 0.1
                    else:
                        eta = 0.0
                    label = mixup_transform(label, classes, lam, eta)

                elif self._cfg.train.label_smoothing:
                    hard_label = label
                    label = smooth(label, self.num_class)

                if self.distillation:
                    teacher_prob = [nd.softmax(self.teacher(X.astype(self._cfg.train.dtype, copy=False)) \
                                    / self._cfg.train.temperature) for X in data]

                with ag.record():
                    outputs = [self.net(X.astype(self._cfg.train.dtype, copy=False)) for X in data]
                    if self.distillation:
                        losses = [L(yhat.astype('float32', copy=False),
                                    y.astype('float32', copy=False),
                                    p.astype('float32', copy=False)) \
                                        for yhat, y, p in zip(outputs, label, teacher_prob)]
                    else:
                        losses = [L(yhat,
                                    y.astype(self._cfg.train.dtype, copy=False)) for yhat, y in zip(outputs, label)]
                for l in losses:
                    l.backward()
                self.trainer.step(self.batch_size)

                if self._cfg.train.mixup:
                    output_softmax = [nd.SoftmaxActivation(out.astype('float32', copy=False)) \
                                    for out in outputs]
                    train_metric.update(label, output_softmax)
                else:
                    if self._cfg.train.label_smoothing:
                        train_metric.update(hard_label, outputs)
                    else:
                        train_metric.update(label, outputs)

                if self._cfg.train.log_interval and not (i+1)%self._cfg.train.log_interval:
                    train_metric_name, train_metric_score = train_metric.get()
                    self._logger.info('Epoch[%d] Batch [%d]\tSpeed: %f samples/sec\t%s=%f\tlr=%f',
                                      epoch, i,
                                      self._cfg.train.batch_size*self._cfg.train.log_interval/(time.time()-last_tic),
                                      train_metric_name, train_metric_score, self.trainer.learning_rate)
                    last_tic = time.time()
                self._time_elapsed += time.time() - btic

            post_tic = time.time()
            train_metric_name, train_metric_score = train_metric.get()
            throughput = int(self.batch_size * i /(time.time() - tic))

            self._logger.info('[Epoch %d] training: %s=%f', epoch, train_metric_name, train_metric_score)
            self._logger.info('[Epoch %d] speed: %d samples/sec\ttime cost: %f', epoch, throughput, time.time()-tic)

            val_score = self._evaluate(val_data, metric_name=train_metric_name)
            if train_metric_name == 'rmse':
                early_stopper.update(-val_score)
                self._logger.info('[Epoch %d] validation: rmse=%f', epoch, val_score)
                if -val_score > self._best_acc:
                    cp_name = os.path.join(self._logdir, _BEST_CHECKPOINT_FILE)
                    self._logger.info('[Epoch %d] Current best rmse: %f vs previous %f, saved to %s',
                                      self.epoch, val_score, -self._best_acc, cp_name)
                    self.save(cp_name)
                    self._best_acc = -val_score
                if self._reporter:
                    self._reporter(epoch=epoch, acc_reward=-val_score)
            else:
                top1_val, top5_val = val_score
                early_stopper.update(top1_val)
                self._logger.info('[Epoch %d] validation: top1=%f top5=%f', epoch, top1_val, top5_val)

                if top1_val > self._best_acc:
                    cp_name = os.path.join(self._logdir, _BEST_CHECKPOINT_FILE)
                    self._logger.info('[Epoch %d] Current best top-1: %f vs previous %f, saved to %s',
                                      self.epoch, top1_val, self._best_acc, cp_name)
                    self.save(cp_name)
                    self._best_acc = top1_val
                if self._reporter:
                    self._reporter(epoch=epoch, acc_reward=top1_val)
            self._time_elapsed += time.time() - post_tic
        if train_metric_name == 'rmse': # TODO: fix ImageClassification(BaseTask)._fit_summary()
            return {'train_score': train_metric_score, 'valid_score': -self._best_acc,
                    'time': self._time_elapsed, 'checkpoint': cp_name}
        else: # accuracy by default:
            return {'train_acc': train_metric_score, 'valid_acc': self._best_acc,
                    'time': self._time_elapsed, 'checkpoint': cp_name}

    def _init_network(self, **kwargs):
        load_only = kwargs.get('load_only', False)
        if not self.num_class and self._problem_type != REGRESSION:
            raise ValueError('This is a classification problem and we are not able to create network when `num_class` is unknown. \
                It should be inferred from dataset or resumed from saved states.')
        assert len(self.classes) == self.num_class
        # ctx
        valid_gpus = []
        if self._cfg.gpus:
            valid_gpus = self._validate_gpus(self._cfg.gpus)
            if not valid_gpus:
                self._logger.warning(
                    'No gpu detected, fallback to cpu. You can ignore this warning if this is intended.')
            elif len(valid_gpus) != len(self._cfg.gpus):
                self._logger.warning(
                    f'Loaded on gpu({valid_gpus}), different from gpu({self._cfg.gpus}).')
        self.ctx = [mx.gpu(int(i)) for i in valid_gpus]
        self.ctx = self.ctx if self.ctx else [mx.cpu()]

        # network
        if self._custom_net is None:
            model_name = self._cfg.img_cls.model.lower()
            input_size = self.input_size
            if 'inception' in model_name or 'googlenet' in model_name:
                self.input_size = 299
            elif 'resnest101' in model_name:
                self.input_size = 256
            elif 'resnest200' in model_name:
                self.input_size = 320
            elif 'resnest269' in model_name:
                self.input_size = 416
            elif 'cifar' in model_name:
                self.input_size = 28
        else:
            self._logger.debug('Custom network specified, ignore the model name in config...')
            self.net = copy.deepcopy(self._custom_net)
            model_name = ''
            self.input_size = input_size = self._cfg.train.input_size


        if input_size != self.input_size:
            self._logger.info(f'Change input size to {self.input_size}, given model type: {model_name}')

        use_pretrained = not load_only and self._cfg.img_cls.use_pretrained
        if self._problem_type == REGRESSION:
            self.num_class = 0
        out_channels = max(self.num_class, 1)
        if use_pretrained:
            kwargs = {'ctx': self.ctx, 'pretrained': True, 'classes': 1000 if 'cifar' not in model_name else 10}
        else:
            kwargs = {'ctx': self.ctx, 'pretrained': False, 'classes': out_channels}
        if self._cfg.img_cls.use_gn:
            kwargs['norm_layer'] = nn.GroupNorm
        if model_name.startswith('vgg'):
            kwargs['batch_norm'] = self._cfg.img_cls.batch_norm
        elif model_name.startswith('resnext'):
            kwargs['use_se'] = self._cfg.img_cls.use_se

        if self._cfg.img_cls.last_gamma:
            kwargs['last_gamma'] = True

        if model_name:
            self.net = get_model(model_name, **kwargs)
        if model_name and use_pretrained:
            # reset last fully connected layer
            fc_layer_found = False
            for fc_name in ('output', 'fc'):
                fc_layer = getattr(self.net, fc_name, None)
                if fc_layer is not None:
                    fc_layer_found = True
                    break
            if fc_layer_found:
                in_channels = list(fc_layer.collect_params().values())[0].shape[1]
                if isinstance(fc_layer, gluon.nn.Dense):
                    new_fc_layer = gluon.nn.Dense(out_channels, in_units=in_channels)
                elif isinstance(fc_layer, gluon.nn.Conv2D):
                    new_fc_layer = gluon.nn.Conv2D(out_channels, in_channels=in_channels, kernel_size=1)
                elif isinstance(fc_layer, gluon.nn.HybridSequential):
                    new_fc_layer = gluon.nn.HybridSequential(prefix='output_')
                    with new_fc_layer.name_scope():
                        new_fc_layer.add(gluon.nn.Conv2D(out_channels, in_channels=in_channels, kernel_size=1))
                        new_fc_layer.add(gluon.nn.Flatten())
                else:
                    raise TypeError(f'Invalid FC layer type {type(fc_layer)} found, expected (Conv2D, Dense)...')
                new_fc_layer.initialize(mx.init.MSRAPrelu(), ctx=self.ctx)
                self.net.collect_params().setattr('lr_mult', self._cfg.train.transfer_lr_mult)
                new_fc_layer.collect_params().setattr('lr_mult', self._cfg.train.output_lr_mult)
                self._logger.debug(f'Reduce network lr multiplier to {self._cfg.train.transfer_lr_mult}, while keep ' +
                                   f'last FC layer lr_mult to {self._cfg.train.output_lr_mult}')
                setattr(self.net, fc_name, new_fc_layer)
            else:
                raise RuntimeError('Unable to modify the last fc layer in network, (output, fc) expected...')
        else:
            self.net.initialize(mx.init.MSRAPrelu(), ctx=self.ctx)
        self.net.cast(self._cfg.train.dtype)

        # teacher model for distillation training
        if not load_only and self._cfg.train.teacher is not None and self._cfg.train.hard_weight < 1.0 and self.num_class == 1000:
            teacher_name = self._cfg.train.teacher
            self.teacher = get_model(teacher_name, pretrained=True, classes=self.num_class, ctx=self.ctx)
            self.teacher.cast(self._cfg.train.dtype)
            self.teacher.collect_params().reset_ctx(self.ctx)
            self.distillation = True
        else:
            self.distillation = False
        self.net.collect_params().reset_ctx(self.ctx)

    def _init_trainer(self):
        if self.last_train is None:
            raise RuntimeError('Cannot init trainer without knowing the size of training data')
        if isinstance(self.last_train, pd.DataFrame):
            train_size = len(self.last_train)
        elif isinstance(self.last_train, int):
            train_size = self.last_train
        else:
            raise ValueError("Unknown type of self.last_train: {}".format(type(self.last_train)))


        num_gpus = len(self.ctx)
        batch_size = self._cfg.train.batch_size
        self.batch_size = batch_size
        lr_decay = self._cfg.train.lr_decay
        lr_decay_period = self._cfg.train.lr_decay_period
        if self._cfg.train.lr_decay_period > 0:
            lr_decay_epoch = list(range(lr_decay_period, self._cfg.train.epochs, lr_decay_period))
        else:
            lr_decay_epoch = [int(i) for i in self._cfg.train.lr_decay_epoch.split(',')]
        lr_decay_epoch = [e - self._cfg.train.warmup_epochs for e in lr_decay_epoch]
        num_batches = train_size // batch_size

        lr_scheduler = LRSequential([
            LRScheduler('linear', base_lr=0, target_lr=self._cfg.train.lr,
                        nepochs=self._cfg.train.warmup_epochs, iters_per_epoch=num_batches),
            LRScheduler(self._cfg.train.lr_mode, base_lr=self._cfg.train.lr, target_lr=0,
                        nepochs=self._cfg.train.epochs - self._cfg.train.warmup_epochs,
                        iters_per_epoch=num_batches,
                        step_epoch=lr_decay_epoch,
                        step_factor=lr_decay, power=2)
        ])

        if self._optimizer is None:
            optimizer = 'nag'
            optimizer_params = {'wd': self._cfg.train.wd,
                                'momentum': self._cfg.train.momentum,
                                'lr_scheduler': lr_scheduler}
            if self._cfg.train.dtype != 'float32':
                optimizer_params['multi_precision'] = True
            self.trainer = gluon.Trainer(self.net.collect_params(), optimizer, optimizer_params)
        else:
            optimizer = self._optimizer
            if isinstance(optimizer, str):
                try:
                    optimizer = mx.optimizer.create(optimizer, learning_rate=self._cfg.train.lr)
                except TypeError:
                    optimizer = mx.optimizer.create(optimizer)
            self.trainer = gluon.Trainer(self.net.collect_params(), optimizer)

    def evaluate(self, val_data, metric_name=None):
        return self._evaluate(val_data, metric_name=metric_name)

    def _evaluate(self, val_data, metric_name=None):
        """Test on validation dataset."""

        if not isinstance(val_data, (gluon.data.DataLoader, mx.io.MXDataIter)):
            if hasattr(val_data, 'to_mxnet'):
                val_data = val_data.to_mxnet()
            resize = int(math.ceil(self.input_size / self._cfg.train.crop_ratio))
            normalize = transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
            transform_test = transforms.Compose([
                transforms.Resize(resize, keep_ratio=True),
                transforms.CenterCrop(self.input_size),
                transforms.ToTensor(),
                normalize
            ])
            val_data = gluon.data.DataLoader(
                val_data.transform_first(transform_test),
                batch_size=self._cfg.valid.batch_size, shuffle=False, last_batch='keep',
                num_workers=self._cfg.valid.num_workers)

        if metric_name == 'rmse':
            rmse_metric = mx.metric.RMSE()
            for _, batch in enumerate(val_data):
                data, label = self.batch_fn(batch, self.ctx)
                outputs = [self.net(X.astype(self._cfg.train.dtype, copy=False)) for X in data]
                rmse_metric.update(label, outputs)

            _, val_score = rmse_metric.get()
            return val_score

        else: # accuracy by default
            acc_top1 = mx.metric.Accuracy()
            acc_top5 = mx.metric.TopKAccuracy(min(5, self.num_class))
            for _, batch in enumerate(val_data):
                data, label = self.batch_fn(batch, self.ctx)
                outputs = [self.net(X.astype(self._cfg.train.dtype, copy=False)) for X in data]
                acc_top1.update(label, outputs)
                acc_top5.update(label, outputs)

            _, top1 = acc_top1.get()
            _, top5 = acc_top5.get()
            return top1, top5

    def _predict_preprocess(self, x):
        resize = int(math.ceil(self.input_size / self._cfg.train.crop_ratio))
        if isinstance(x, str):
            x = self._predict_preprocess(transform_eval(
                mx.image.imread(x), resize_short=resize, crop_size=self.input_size))
        elif isinstance(x, Image.Image):
            x = self._predict_preprocess(np.array(x))
        elif isinstance(x, np.ndarray):
            x = self._predict_preprocess(mx.nd.array(x))
        elif isinstance(x, mx.nd.NDArray):
            if len(x.shape) == 3 and x.shape[-1] == 3:
                x = transform_eval(x, resize_short=resize, crop_size=self.input_size)
            elif len(x.shape) == 4 and x.shape[1] == 3:
                expected = (self.input_size, self.input_size)
                assert x.shape[2:] == expected, "Expected: {}, given {}".format(expected, x.shape[2:])
            elif x.shape[1] == 1:
                # gray image to rgb
                x = mx.nd.concat([x] * 3, dim=1)
            else:
                raise ValueError('array input with shape (h, w, 3) or (n, 3, h, w) is required for predict')
        return x

    def _predict(self, x, ctx_id=0, with_proba=False):
        if with_proba and self._problem_type in [MULTICLASS, BINARY]:
            return self._predict_proba(x, ctx_id=ctx_id)
        x = self._predict_preprocess(x)
        if isinstance(x, pd.DataFrame):
            assert 'image' in x.columns, "Expect column `image` for input images"
            df = self._predict(tuple(x['image']))
            return df.reset_index(drop=True)
        elif isinstance(x, (list, tuple)):
            bs = self._cfg.valid.batch_size
            self.net.hybridize()
            results = []
            topK = min(5, self.num_class)
            loader = mx.gluon.data.DataLoader(
                ImageListDataset(x, self._predict_preprocess), batch_size=bs, last_batch='keep')
            idx = 0
            for batch in loader:
                batch = mx.gluon.utils.split_and_load(batch, ctx_list=self.ctx, even_split=False)
                pred = [self.net(input) for input in batch]
                for p in pred:
                    for ii in range(p.shape[0]):
                        if self._problem_type in [MULTICLASS, BINARY]:
                            ind = nd.topk(p[ii], k=topK).astype('int').asnumpy().flatten()
                            probs = mx.nd.softmax(p[ii]).asnumpy().flatten()
                            results.extend([{'class': self.classes[ind[k]],
                                             'score': probs[ind[k]], 'id': ind[k], 'image': x[idx]}
                                            for k in range(topK)])
                        else:
                            results.append({'prediction': p[ii].asnumpy().flatten()[0], 'image': x[idx]})
                        idx += 1
            return pd.DataFrame(results)
        elif not isinstance(x, mx.nd.NDArray):
            raise ValueError('Input is not supported: {}'.format(type(x)))
        assert len(x.shape) == 4 and x.shape[1] == 3, "Expect input to be (n, 3, h, w), given {}".format(x.shape)
        x = x.as_in_context(self.ctx[ctx_id])
        pred = self.net(x)
        topK = min(5, self.num_class)
        if self._problem_type in [MULTICLASS, BINARY]:
            ind = nd.topk(pred, k=topK)[0].astype('int').asnumpy().flatten()
            probs = mx.nd.softmax(pred)[0].asnumpy().flatten()
            df = pd.DataFrame([{'class': self.classes[ind[i]], 'score': probs[ind[i]], 'id': ind[i]} for i in range(topK)])
        else:
            df = pd.DataFrame([{'prediction': pred.asnumpy().flatten()[0]}])
        return df

    def _get_feature_net(self):
        """Get the network slice for feature extraction only"""
        if hasattr(self, '_feature_net') and self._feature_net is not None:
            return self._feature_net
        self._feature_net = copy.copy(self.net)
        fc_layer_found = False
        for fc_name in ('output', 'fc'):
            fc_layer = getattr(self._feature_net, fc_name, None)
            if fc_layer is not None:
                fc_layer_found = True
                break
        if fc_layer_found:
            self._feature_net.register_child(nn.Identity(), fc_name)
            super(gluon.Block, self._feature_net).__setattr__(fc_name, nn.Identity())
            self.net.__setattr__(fc_name, fc_layer)
        else:
            raise RuntimeError('Unable to modify the last fc layer in network, (output, fc) expected...')
        return self._feature_net

    def _predict_feature(self, x, ctx_id=0):
        x = self._predict_preprocess(x)
        if isinstance(x, pd.DataFrame):
            assert 'image' in x.columns, "Expect column `image` for input images"
            df = self._predict_feature(tuple(x['image']))
            df = df.set_index(x.index)
            df['image'] = x['image']
            return df
        elif isinstance(x, (list, tuple)):
            assert isinstance(x[0], str), "expect image paths in list/tuple input"
            bs = self._cfg.valid.batch_size
            feat_net = self._get_feature_net()
            feat_net.hybridize()
            results = []
            loader = mx.gluon.data.DataLoader(
                ImageListDataset(x, self._predict_preprocess), batch_size=bs, last_batch='keep')
            for batch in loader:
                batch = mx.gluon.utils.split_and_load(batch, ctx_list=self.ctx, even_split=False)
                feats = [feat_net(input) for input in batch]
                for p in feats:
                    for ii in range(p.shape[0]):
                        feat = p[ii].asnumpy().flatten()
                        results.append({'image_feature': feat})
            df = pd.DataFrame(results)
            df['image'] = x
            return df
        elif not isinstance(x, mx.nd.NDArray):
            raise ValueError('Input is not supported: {}'.format(type(x)))
        assert len(x.shape) == 4 and x.shape[1] == 3, "Expect input to be (n, 3, h, w), given {}".format(x.shape)
        x = x.as_in_context(self.ctx[ctx_id])
        feat_net = self._get_feature_net()
        results = []
        for ii in range(x.shape[0]):
            feat = feat_net(x)[ii].asnumpy().flatten()
            results.append({'image_feature': feat})
        df = pd.DataFrame(results)
        return df

    def _predict_proba(self, x, ctx_id=0):
        if self._problem_type not in [MULTICLASS, BINARY]:
            raise AssertionError('predict_proba is only supported for classification problems. Please use predict instead.')
        x = self._predict_preprocess(x)
        if isinstance(x, pd.DataFrame):
            assert 'image' in x.columns, "Expect column `image` for input images"
            df = self._predict_proba(tuple(x['image']))
            return df.reset_index(drop=True)
        elif isinstance(x, (list, tuple)):
            bs = self._cfg.valid.batch_size
            self.net.hybridize()
            results = []
            loader = mx.gluon.data.DataLoader(
                ImageListDataset(x, self._predict_preprocess), batch_size=bs, last_batch='keep')
            idx = 0
            for batch in loader:
                batch = mx.gluon.utils.split_and_load(batch, ctx_list=self.ctx, even_split=False)
                pred = [self.net(input) for input in batch]
                for p in pred:
                    probs = mx.nd.softmax(p, axis=-1)
                    for ii in range(p.shape[0]):
                        prob = probs[ii]
                        results.append({'image_proba': prob.asnumpy().flatten().tolist(), 'image': x[idx]})
                        idx += 1
            return pd.DataFrame(results)
        elif not isinstance(x, mx.nd.NDArray):
            raise ValueError('Input is not supported: {}'.format(type(x)))
        assert len(x.shape) == 4 and x.shape[1] == 3, "Expect input to be (n, 3, h, w), given {}".format(x.shape)
        x = x.as_in_context(self.ctx[ctx_id])
        pred = self.net(x)
        probs = mx.nd.softmax(pred)[0].asnumpy().flatten().tolist()
        df = pd.DataFrame([{'image_proba': probs}])
        return df

class ImageListDataset(mx.gluon.data.Dataset):
    """An internal image list dataset for batch predict"""
    def __init__(self, imlist, fn):
        self._imlist = imlist
        self._fn = fn

    def __getitem__(self, idx):
        img = self._fn(self._imlist[idx])[0]
        return img

    def __len__(self):
        return len(self._imlist)



"""CenterNet Estimator"""
# pylint: disable=unused-variable,missing-function-docstring,abstract-method,logging-format-interpolation,arguments-differ,logging-not-lazy
import os
import math
import time
import warnings
from collections import OrderedDict

from PIL import Image
import pandas as pd
import numpy as np
import mxnet as mx
from mxnet import autograd
from mxnet import gluon

from ..base_estimator import BaseEstimator, set_default
from ....data.batchify import Tuple, Stack, Pad
from ....data.transforms.presets.center_net import CenterNetDefaultTrainTransform
from ....data.transforms.presets.center_net import CenterNetDefaultValTransform
from ....data.transforms.presets.center_net import load_test, transform_test
from ....loss import MaskedL1Loss, HeatmapFocalLoss
from ....model_zoo import get_model
from ....model_zoo.center_net import get_center_net, get_base_network
from ....utils import LRScheduler, LRSequential
from ....utils.metrics import VOCMApMetric, VOC07MApMetric
from .default import CenterNetCfg
from ...data.dataset import ObjectDetectionDataset
from ..conf import _BEST_CHECKPOINT_FILE
from ..utils import EarlyStopperOnPlateau

__all__ = ['CenterNetEstimator']

@set_default(CenterNetCfg())
class CenterNetEstimator(BaseEstimator):
    """Estimator implementation for CenterNet.

    Parameters
    ----------
    config : dict
        Config in nested dict.
    logger : logging.Logger
        Optional logger for this estimator, can be `None` when default setting is used.
    reporter : callable
        The reporter for metric checkpointing.

    """
    Dataset = ObjectDetectionDataset
    def __init__(self, config, logger=None, reporter=None):
        super(CenterNetEstimator, self).__init__(config, logger, reporter=reporter, name=None)
        self.last_train = None

    def _predict(self, x, ctx_id=0):
        short_size = min(self._cfg.center_net.data_shape)
        if isinstance(x, str):
            x = load_test(x, short=short_size, max_size=1024)[0]
        elif isinstance(x, Image.Image):
            return self._predict(np.array(x))
        elif isinstance(x, np.ndarray):
            return self._predict(mx.nd.array(x))
        elif isinstance(x, mx.nd.NDArray):
            if len(x.shape) != 3 or x.shape[-1] != 3:
                raise ValueError('array input with shape (h, w, 3) is required for predict')
            x = transform_test(x, short=short_size, max_size=1024)[0]
        elif isinstance(x, pd.DataFrame):
            assert 'image' in x.columns, "Expect column `image` for input images"
            def _predict_merge(x, ctx_id=0):
                y = self._predict(x, ctx_id=ctx_id)
                y['image'] = x
                return y
            return pd.concat([_predict_merge(xx, ctx_id=ii % len(self.ctx)) \
                for ii, xx in enumerate(x['image'])]).reset_index(drop=True)
        elif isinstance(x, (list, tuple)):
            return pd.concat([self._predict(xx, ctx_id=ii % len(self.ctx)) \
                for ii, xx in enumerate(x)]).reset_index(drop=True)
        else:
            raise ValueError('Input is not supported: {}'.format(type(x)))
        height, width = x.shape[2:4]
        x = x.as_in_context(self.ctx[ctx_id])
        ids, scores, bboxes = [xx[0].asnumpy() for xx in self.net(x)]
        bboxes[:, (0, 2)] /= width
        bboxes[:, (1, 3)] /= height
        bboxes = np.clip(bboxes, 0.0, 1.0).tolist()
        df = pd.DataFrame({'predict_class': [self.classes[int(id)] for id in ids], 'predict_score': scores,
                           'predict_rois': [{'xmin': bbox[0], 'ymin': bbox[1], 'xmax': bbox[2], 'ymax': bbox[3]} \
                                for bbox in bboxes]})
        # filter out invalid (scores < 0) rows
        valid_df = df[df['predict_score'] > 0].reset_index(drop=True)
        return valid_df

    def _fit(self, train_data, val_data, time_limit=math.inf):
        tic = time.time()
        self._best_map = 0
        self.epoch = 0
        self._time_elapsed = 0
        if max(self._cfg.train.start_epoch, self.epoch) >= self._cfg.train.epochs:
            return {'time', self._time_elapsed}
        if not isinstance(train_data, pd.DataFrame):
            self.last_train = len(train_data)
        else:
            self.last_train = train_data
        self._init_trainer()
        self._time_elapsed += time.time() - tic
        return self._resume_fit(train_data, val_data, time_limit=time_limit)

    def _resume_fit(self, train_data, val_data, time_limit=math.inf):
        tic = time.time()
        if max(self._cfg.train.start_epoch, self.epoch) >= self._cfg.train.epochs:
            return {'time', self._time_elapsed}
        if not self.classes or not self.num_class:
            raise ValueError('Unable to determine classes of dataset')
        train_dataset = train_data.to_mxnet()
        val_dataset = val_data.to_mxnet()

        # dataloader
        batch_size = self._cfg.train.batch_size
        width, height = self._cfg.center_net.data_shape
        num_class = len(train_dataset.classes)
        batchify_fn = Tuple([Stack() for _ in range(6)])  # stack image, cls_targets, box_targets
        train_loader = gluon.data.DataLoader(
            train_dataset.transform(CenterNetDefaultTrainTransform(
                width, height, num_class=num_class, scale_factor=self.net.scale)),
            batch_size, True, batchify_fn=batchify_fn, last_batch='rollover',
            num_workers=self._cfg.train.num_workers)
        val_batchify_fn = Tuple(Stack(), Pad(pad_val=-1))
        val_loader = gluon.data.DataLoader(
            val_dataset.transform(CenterNetDefaultValTransform(width, height)),
            self._cfg.valid.batch_size, False, batchify_fn=val_batchify_fn, last_batch='keep',
            num_workers=self._cfg.valid.num_workers)
        train_eval_loader = gluon.data.DataLoader(
            train_dataset.transform(CenterNetDefaultValTransform(width, height)),
            self._cfg.valid.batch_size, False, batchify_fn=val_batchify_fn, last_batch='keep',
            num_workers=self._cfg.valid.num_workers)
        self._time_elapsed += time.time() - tic
        return self._train_loop(train_loader, val_loader, train_eval_loader, time_limit=time_limit)

    def _train_loop(self, train_data, val_data, train_eval_data, time_limit=math.inf):
        start_tic = time.time()
        wh_loss = MaskedL1Loss(weight=self._cfg.center_net.wh_weight)
        heatmap_loss = HeatmapFocalLoss(from_logits=True)
        center_reg_loss = MaskedL1Loss(weight=self._cfg.center_net.center_reg_weight)
        heatmap_loss_metric = mx.metric.Loss('HeatmapFocal')
        wh_metric = mx.metric.Loss('WHL1')
        center_reg_metric = mx.metric.Loss('CenterRegL1')

        self._logger.info('Start training from [Epoch %d]', max(self._cfg.train.start_epoch, self.epoch))
        early_stopper = EarlyStopperOnPlateau(
            patience=self._cfg.train.early_stop_patience,
            min_delta=self._cfg.train.early_stop_min_delta,
            baseline_value=self._cfg.train.early_stop_baseline,
            max_value=self._cfg.train.early_stop_max_value)
        mean_ap = [-1]
        cp_name = ''
        self._time_elapsed += time.time() - start_tic
        for self.epoch in range(max(self._cfg.train.start_epoch, self.epoch), self._cfg.train.epochs):
            epoch = self.epoch
            tic = time.time()
            last_tic = time.time()
            if self._best_map >= 1.0:
                self._logger.info('[Epoch %d] Early stopping as mAP is reaching 1.0', epoch)
                break
            should_stop, stop_message = early_stopper.get_early_stop_advice()
            if should_stop:
                self._logger.info('[Epoch {}] '.format(epoch) + stop_message)
                break
            wh_metric.reset()
            center_reg_metric.reset()
            heatmap_loss_metric.reset()
            self.net.hybridize()

            for i, batch in enumerate(train_data):
                btic = time.time()
                if self._time_elapsed > time_limit:
                    self._logger.warning(f'`time_limit={time_limit}` reached, exit early...')
                    return {'train_map': float(mean_ap[-1]), 'valid_map': self._best_map,
                            'time': self._time_elapsed, 'checkpoint': cp_name}
                split_data = [
                    gluon.utils.split_and_load(batch[ind], ctx_list=self.ctx, batch_axis=0, even_split=False) for ind
                    in range(6)]
                data, heatmap_targets, wh_targets, wh_masks, center_reg_targets, center_reg_masks = split_data
                batch_size = self._cfg.train.batch_size
                with autograd.record():
                    sum_losses = []
                    heatmap_losses = []
                    wh_losses = []
                    center_reg_losses = []
                    wh_preds = []
                    center_reg_preds = []
                    for x, heatmap_target, wh_target, wh_mask, center_reg_target, center_reg_mask in zip(
                            *split_data):
                        heatmap_pred, wh_pred, center_reg_pred = self.net(x)
                        wh_preds.append(wh_pred)
                        center_reg_preds.append(center_reg_pred)
                        wh_losses.append(wh_loss(wh_pred, wh_target, wh_mask))
                        center_reg_losses.append(
                            center_reg_loss(center_reg_pred, center_reg_target, center_reg_mask))
                        heatmap_losses.append(heatmap_loss(heatmap_pred, heatmap_target))
                        curr_loss = heatmap_losses[-1] + wh_losses[-1] + center_reg_losses[-1]
                        sum_losses.append(curr_loss)
                    autograd.backward(sum_losses)
                self.trainer.step(len(sum_losses))  # step with # gpus

                heatmap_loss_metric.update(0, heatmap_losses)
                wh_metric.update(0, wh_losses)
                center_reg_metric.update(0, center_reg_losses)
                if self._cfg.train.log_interval and not (i + 1) % self._cfg.train.log_interval:
                    name2, loss2 = wh_metric.get()
                    name3, loss3 = center_reg_metric.get()
                    name4, loss4 = heatmap_loss_metric.get()
                    self._logger.info(
                        '[Epoch {}][Batch {}], Speed: {:.3f} samples/sec, '
                        'LR={}, {}={:.3f}, {}={:.3f}, {}={:.3f}'.format(
                            epoch, i, batch_size / (time.time() - last_tic),
                            self.trainer.learning_rate, name2, loss2, name3, loss3, name4, loss4))
                    last_tic = time.time()
                self._time_elapsed += time.time() - btic

            post_tic = time.time()
            name2, loss2 = wh_metric.get()
            name3, loss3 = center_reg_metric.get()
            name4, loss4 = heatmap_loss_metric.get()
            self._logger.info(
                '[Epoch {}] Training cost: {:.3f}, {}={:.3f}, {}={:.3f}, {}={:.3f}'.format(
                    epoch, (time.time() - tic), name2, loss2, name3, loss3, name4, loss4))
            if (epoch % self._cfg.valid.interval == 0) or (epoch == self._cfg.train.epochs - 1):
                # consider reduce the frequency of validation to save time
                map_name, mean_ap = self._evaluate(val_data)
                val_msg = '\n'.join(['{}={}'.format(k, v) for k, v in zip(map_name, mean_ap)])
                self._logger.info('[Epoch %d] Validation: \n%s', epoch, val_msg)
                current_map = float(mean_ap[-1])
                if current_map > self._best_map:
                    cp_name = os.path.join(self._logdir, _BEST_CHECKPOINT_FILE)
                    self._logger.info('[Epoch %d] Current best map: %f vs previous %f, saved to %s',
                                      self.epoch, current_map, self._best_map, cp_name)
                    self.save(cp_name)
                    self._best_map = current_map
                if self._reporter:
                    self._reporter(epoch=epoch, map_reward=current_map)
                early_stopper.update(current_map, epoch=epoch)
            self._time_elapsed += time.time() - post_tic
        # map on train data
        tic = time.time()
        map_name, mean_ap = self._evaluate(train_eval_data)
        self._time_elapsed += time.time() - tic
        return {'train_map': float(mean_ap[-1]), 'valid_map': self._best_map,
                'time': self._time_elapsed, 'checkpoint': cp_name}

    def _evaluate(self, val_data):
        """Test on validation dataset."""
        if self._cfg.valid.metric == 'voc07':
            eval_metric = VOC07MApMetric(iou_thresh=self._cfg.valid.iou_thresh, class_names=self.classes)
        elif self._cfg.valid.metric == 'voc':
            eval_metric = VOCMApMetric(iou_thresh=self._cfg.valid.iou_thresh, class_names=self.classes)
        else:
            raise ValueError(f'Invalid metric type: {self._cfg.valid.metric}')
        self.net.flip_test = self._cfg.valid.flip_test
        mx.nd.waitall()
        self.net.hybridize()
        if not isinstance(val_data, gluon.data.DataLoader):
            if hasattr(val_data, 'to_mxnet'):
                val_data = val_data.to_mxnet()
            val_batchify_fn = Tuple(Stack(), Pad(pad_val=-1))
            width, height = self._cfg.center_net.data_shape
            val_data = gluon.data.DataLoader(
                val_data.transform(CenterNetDefaultValTransform(width, height)),
                self._cfg.valid.batch_size, False, batchify_fn=val_batchify_fn, last_batch='keep',
                num_workers=self._cfg.valid.num_workers)
        for batch in val_data:
            val_ctx = self.ctx
            if batch[0].shape[0] < len(val_ctx):
                val_ctx = val_ctx[:batch[0].shape[0]]
            data = gluon.utils.split_and_load(batch[0], ctx_list=val_ctx, batch_axis=0,
                                              even_split=False)
            label = gluon.utils.split_and_load(batch[1], ctx_list=val_ctx, batch_axis=0,
                                               even_split=False)
            det_bboxes = []
            det_ids = []
            det_scores = []
            gt_bboxes = []
            gt_ids = []
            gt_difficults = []
            for x, y in zip(data, label):
                # get prediction results
                ids, scores, bboxes = self.net(x)
                det_ids.append(ids)
                det_scores.append(scores)
                # clip to image size
                det_bboxes.append(bboxes.clip(0, batch[0].shape[2]))
                # split ground truths
                gt_ids.append(y.slice_axis(axis=-1, begin=4, end=5))
                gt_bboxes.append(y.slice_axis(axis=-1, begin=0, end=4))
                gt_difficults.append(
                    y.slice_axis(axis=-1, begin=5, end=6) if y.shape[-1] > 5 else None)

            # update metric
            eval_metric.update(det_bboxes, det_ids, det_scores, gt_bboxes, gt_ids, gt_difficults)
        return eval_metric.get()

    def _init_network(self, **kwargs):
        load_only = kwargs.get('load_only', False)
        if not self.num_class:
            raise ValueError('Unable to create network when `num_class` is unknown. \
                It should be inferred from dataset or resumed from saved states.')
        assert len(self.classes) == self.num_class
        # network
        valid_gpus = []
        if self._cfg.gpus:
            valid_gpus = self._validate_gpus(self._cfg.gpus)
            if not valid_gpus:
                self._logger.warning(
                    'No gpu detected, fallback to cpu. You can ignore this warning if this is intended.')
            elif len(valid_gpus) != len(self._cfg.gpus):
                self._logger.warning(
                    f'Loaded on gpu({valid_gpus}), different from gpu({self._cfg.gpus}).')
        ctx = [mx.gpu(int(i)) for i in valid_gpus]
        ctx = ctx if ctx else [mx.cpu()]
        self.ctx = ctx
        if self._cfg.center_net.transfer is not None:
            assert isinstance(self._cfg.center_net.transfer, str)
            self._logger.info('Using transfer learning from %s, ignoring some of the network configs',
                              self._cfg.center_net.transfer)
            net = get_model(self._cfg.center_net.transfer, pretrained=(not load_only))
            if load_only:
                with warnings.catch_warnings(record=True) as w:
                    warnings.simplefilter("always")
                    net.initialize()
                net(mx.nd.zeros((1, 3, self._cfg.center_net.data_shape[0], self._cfg.center_net.data_shape[1])))
            net.reset_class(self.classes, reuse_weights=[cname for cname in self.classes if cname in net.classes])
        else:
            net_name = '_'.join(('center_net', self._cfg.center_net.base_network, self.dataset))
            heads = OrderedDict([
                ('heatmap',
                 {'num_output': self.num_class, 'bias': self._cfg.center_net.heads.bias}),
                ('wh', {'num_output': self._cfg.center_net.heads.wh_outputs}),
                ('reg', {'num_output': self._cfg.center_net.heads.reg_outputs})])
            base_network = get_base_network(self._cfg.center_net.base_network,
                                            pretrained=self._cfg.train.pretrained_base and not load_only)
            net = get_center_net(self._cfg.center_net.base_network,
                                 self.dataset,
                                 base_network=base_network,
                                 heads=heads,
                                 head_conv_channel=self._cfg.center_net.heads.head_conv_channel,
                                 classes=self.classes,
                                 scale=self._cfg.center_net.scale,
                                 topk=self._cfg.center_net.topk,
                                 norm_layer=gluon.nn.BatchNorm)
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                net.initialize()
        self.net = net
        for k, v in self.net.collect_params('.*bias').items():
            v.wd_mult = 0.0
        self.net.collect_params().reset_ctx(self.ctx)

    def _init_trainer(self):
        if self.last_train is None:
            raise RuntimeError('Cannot init trainer without knowing the size of training data')
        if isinstance(self.last_train, pd.DataFrame):
            train_size = len(self.last_train)
        elif isinstance(self.last_train, int):
            train_size = self.last_train
        else:
            raise ValueError("Unknown type of self.last_train: {}".format(type(self.last_train)))
        lr_decay = float(self._cfg.train.lr_decay)
        lr_steps = sorted(self._cfg.train.lr_decay_epoch)
        lr_decay_epoch = [e - self._cfg.train.warmup_epochs for e in lr_steps]
        num_batches = train_size // self._cfg.train.batch_size
        lr_scheduler = LRSequential([
            LRScheduler('linear', base_lr=0, target_lr=self._cfg.train.lr,
                        nepochs=self._cfg.train.warmup_epochs, iters_per_epoch=num_batches),
            LRScheduler(self._cfg.train.lr_mode, base_lr=self._cfg.train.lr,
                        nepochs=self._cfg.train.epochs - self._cfg.train.warmup_epochs,
                        iters_per_epoch=num_batches,
                        step_epoch=lr_decay_epoch,
                        step_factor=self._cfg.train.lr_decay, power=2),
        ])
        self.trainer = gluon.Trainer(
            self.net.collect_params(), 'adam',
            {'learning_rate': self._cfg.train.lr, 'wd': self._cfg.train.wd,
             'lr_scheduler': lr_scheduler})



"""Dual Attention Network
https://arxiv.org/abs/1809.02983"""
from mxnet.gluon import nn
from mxnet.context import cpu
from mxnet.gluon.nn import HybridBlock
from .segbase import SegBaseModel
from .fcn import _FCNHead
from .attention import PAM_Module, CAM_Module
# pylint: disable-all

__all__ = ['DANet', 'get_danet', 'get_danet_resnet50_citys', 'get_danet_resnet101_citys']


class DANet(SegBaseModel):
    r"""Dual Attention Networks for Semantic Segmentation

    Parameters
    ----------
    nclass : int
        Number of categories for the training dataset.
    backbone : string
        Pre-trained dilated backbone network type (default:'resnet50'; 'resnet50',
        'resnet101').
    norm_layer : object
        Normalization layer used in backbone network (default: :class:`mxnet.gluon.nn.BatchNorm`;

    Reference:
        Jun Fu, Jing Liu, Haijie Tian, Yong Li, Yongjun Bao, Zhiwei Fang, Hanqing Lu. "Dual Attention
        Network for Scene Segmentation." *CVPR*, 2019
    """

    def __init__(self, nclass, backbone='resnet50', aux=False, ctx=cpu(), pretrained_base=True,
                 height=None, width=None, base_size=520, crop_size=480, dilated=True, **kwargs):
        super(DANet, self).__init__(nclass, aux, backbone, ctx=ctx, base_size=base_size,
                                    crop_size=crop_size, pretrained_base=pretrained_base, **kwargs)
        self.aux = aux
        height = height if height is not None else crop_size
        width = width if width is not None else crop_size

        with self.name_scope():
            self.head = DANetHead(2048, nclass, **kwargs)
            self.head.initialize(ctx=ctx)

        self._up_kwargs = {'height': height, 'width': width}

    def hybrid_forward(self, F, x):
        c3, c4 = self.base_forward(x)

        x = self.head(c4)
        x = list(x)
        x[0] = F.contrib.BilinearResize2D(x[0], **self._up_kwargs)
        x[1] = F.contrib.BilinearResize2D(x[1], **self._up_kwargs)
        x[2] = F.contrib.BilinearResize2D(x[2], **self._up_kwargs)

        outputs = [x[0]]
        outputs.append(x[1])
        outputs.append(x[2])

        return tuple(outputs)


class DANetHead(HybridBlock):
    def __init__(self, in_channels, out_channels, norm_layer=nn.BatchNorm, norm_kwargs=None, **kwargs):
        super(DANetHead, self).__init__()
        inter_channels = in_channels // 4
        self.conv5a = nn.HybridSequential()
        self.conv5a.add(nn.Conv2D(in_channels=in_channels, channels=inter_channels, kernel_size=3, 
                                  padding=1, use_bias=False))
        self.conv5a.add(norm_layer(in_channels=inter_channels, **({} if norm_kwargs is None else norm_kwargs)))
        self.conv5a.add(nn.Activation('relu'))

        self.conv5c = nn.HybridSequential()
        self.conv5c.add(nn.Conv2D(in_channels=in_channels, channels=inter_channels, kernel_size=3, 
                                  padding=1, use_bias=False))
        self.conv5c.add(norm_layer(in_channels=inter_channels, **({} if norm_kwargs is None else norm_kwargs)))
        self.conv5c.add(nn.Activation('relu'))

        self.sa = PAM_Module(inter_channels)
        self.sc = CAM_Module(inter_channels)
        self.conv51 = nn.HybridSequential()
        self.conv51.add(nn.Conv2D(in_channels=inter_channels, channels=inter_channels, kernel_size=3, 
                                  padding=1, use_bias=False))
        self.conv51.add(norm_layer(in_channels=inter_channels, **({} if norm_kwargs is None else norm_kwargs)))
        self.conv51.add(nn.Activation('relu'))

        self.conv52 = nn.HybridSequential()
        self.conv52.add(nn.Conv2D(in_channels=inter_channels, channels=inter_channels, kernel_size=3, 
                                  padding=1, use_bias=False))
        self.conv52.add(norm_layer(in_channels=inter_channels, **({} if norm_kwargs is None else norm_kwargs)))
        self.conv52.add(nn.Activation('relu'))

        self.conv6 = nn.HybridSequential()
        # self.conv6.add(nn.Dropout(0.1))
        self.conv6.add(nn.Conv2D(in_channels=512, channels=out_channels, kernel_size=1))

        self.conv7 = nn.HybridSequential()
        # self.conv7.add(nn.Dropout(0.1))
        self.conv7.add(nn.Conv2D(in_channels=512, channels=out_channels, kernel_size=1))

        self.conv8 = nn.HybridSequential()
        # self.conv8.add(nn.Dropout(0.1))
        self.conv8.add(nn.Conv2D(in_channels=512, channels=out_channels, kernel_size=1))

    def hybrid_forward(self, F, x):
        feat1 = self.conv5a(x)
        sa_feat = self.sa(feat1)
        sa_conv = self.conv51(sa_feat)
        sa_output = self.conv6(sa_conv)

        feat2 = self.conv5c(x)
        sc_feat = self.sc(feat2)
        sc_conv = self.conv52(sc_feat)
        sc_output = self.conv7(sc_conv)

        feat_sum = sa_conv + sc_conv
        sasc_output = self.conv8(feat_sum)

        output = [sasc_output]
        output.append(sa_output)
        output.append(sc_output)

        return tuple(output)

    def predict(self, x):
        h, w = x.shape[2:]
        self._up_kwargs['height'] = h
        self._up_kwargs['width'] = w
        _, c4 = self.base_forward(x)
        x = self.head.demo(c4)
        import mxnet.ndarray as F
        pred = F.contrib.BilinearResize2D(x, **self._up_kwargs)
        return pred


def get_danet(dataset='pascal_voc', backbone='resnet50', pretrained=False,
              root='~/.mxnet/models', ctx=cpu(0), **kwargs):
    r"""DANet model from the paper `"Dual Attention Network for Scene Segmentation"
    <https://arxiv.org/abs/1809.02983>`
    """
    acronyms = {
        'pascal_voc': 'voc',
        'pascal_aug': 'voc',
        'ade20k': 'ade',
        'coco': 'coco',
        'citys': 'citys',
    }
    from ..data import datasets
    # infer number of classes
    model = DANet(nclass=datasets[dataset].NUM_CLASS, backbone=backbone, ctx=ctx, **kwargs)
    model.classes = datasets[dataset].classes
    if pretrained:
        from .model_store import get_model_file
        model.load_parameters(get_model_file('danet_%s_%s' % (backbone, acronyms[dataset]),
                                             tag=pretrained, root=root), ctx=ctx)
    return model


def get_danet_resnet50_citys(**kwargs):
    r"""DANet
    Parameters
    ----------
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.
    Examples
    --------
    >>> model = get_danet_resnet50_citys(pretrained=True)
    >>> print(model)
    """
    return get_danet('citys', 'resnet50', **kwargs)


def get_danet_resnet101_citys(**kwargs):
    r"""DANet
    Parameters
    ----------
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.

    Examples
    --------
    >>> model = get_danet_resnet101_citys(pretrained=True)
    >>> print(model)
    """
    return get_danet('citys', 'resnet101', **kwargs)



from __future__ import print_function
from __future__ import division

import os.path as osp
import mxnet as mx
import numpy as np
from mxnet import autograd, gluon
import gluoncv as gcv
from gluoncv.data import transforms
from gluoncv.data import batchify
from gluoncv.data.batchify import Tuple, Stack, Pad
from gluoncv.data.transforms.presets import ssd
from gluoncv.data.transforms.presets import rcnn
from gluoncv.data.transforms.presets import yolo
from gluoncv.data.transforms.presets import center_net
from .tiny_datasets import COCODetectionTiny, COCOInstanceTiny, VOCDetectionTiny, VOCSegmentationTiny

def test_bbox_crop():
    bbox = np.array([[10, 20, 200, 500], [150, 200, 400, 300]])
    np.testing.assert_allclose(transforms.bbox.crop(bbox, None), bbox)
    np.testing.assert_allclose(
        transforms.bbox.crop(bbox, (20, 30, 200, 200), allow_outside_center=True),
        np.array([[  0,   0, 180, 200], [130, 170, 200, 200]]))
    np.testing.assert_allclose(
        transforms.bbox.crop(bbox, (20, 30, 200, 300), allow_outside_center=False),
        np.array([[  0,   0, 180, 300]]))

def test_bbox_flip():
    bbox = np.array([[10, 20, 200, 500], [150, 200, 400, 300]])
    size = (500, 1000)
    np.testing.assert_allclose(
        transforms.bbox.flip(bbox, size, False, False),
        bbox)
    np.testing.assert_allclose(
        transforms.bbox.flip(bbox, size, False, True),
        np.array([[ 10, 500, 200, 980], [150, 700, 400, 800]]))
    np.testing.assert_allclose(
        transforms.bbox.flip(bbox, size, True, False),
        np.array([[300,  20, 490, 500], [100, 200, 350, 300]]))
    np.testing.assert_allclose(
        transforms.bbox.flip(bbox, size, True, True),
        np.array([[300, 500, 490, 980], [100, 700, 350, 800]]))

def test_bbox_resize():
    bbox = np.array([[10, 20, 200, 500], [150, 200, 400, 300]], dtype=np.float32)
    in_size = (600, 1000)
    out_size = (200, 300)
    np.testing.assert_allclose(
        transforms.bbox.resize(bbox, in_size, out_size),
        np.array([[  3.333333,   6.,  66.66667 , 150.], [ 50.,  60.000004, 133.33334 ,  90.]]),
        rtol=1e-3)

def test_bbox_translate():
    bbox = np.array([[10, 20, 200, 500], [150, 200, 400, 300]], dtype=np.float32)
    xoff = np.random.randint(-100, 100)
    yoff = np.random.randint(-100, 100)
    expected = bbox.copy()
    expected[:, (0, 2)] += xoff
    expected[:, (1, 3)] += yoff
    np.testing.assert_allclose(transforms.bbox.translate(bbox, xoff, yoff), expected)

def test_image_imresize():
    image = mx.random.normal(shape=(240, 480, 3)).astype(np.uint8)
    out = transforms.image.imresize(image, 300, 300)
    np.testing.assert_allclose(out.shape, (300, 300, 3))

def test_image_resize_long():
    image = mx.random.normal(shape=(240, 480, 3)).astype(np.uint8)
    out = transforms.image.resize_long(image, 300)
    np.testing.assert_allclose(out.shape, (150, 300, 3))

def test_image_random_pca():
    image = mx.random.normal(shape=(240, 120, 3)).astype(np.float32)
    out = transforms.image.random_pca_lighting(image, 0.1)
    np.testing.assert_allclose(out.shape, image.shape)  # no value check

def test_image_random_expand():
    image = mx.random.normal(shape=(240, 120, 3)).astype(np.uint8)
    # no expand when ratio <= 1
    out, _ = transforms.image.random_expand(image, max_ratio=0.1, keep_ratio=True)
    np.testing.assert_allclose(out.asnumpy(), image.asnumpy())
    # check ratio
    out, _ = transforms.image.random_expand(image, 4, keep_ratio=True)
    np.testing.assert_allclose(out.shape[0] / out.shape[1], image.shape[0] / image.shape[1], rtol=1e-2, atol=1e-3)
    # #
    out, _ = transforms.image.random_expand(image, 4, keep_ratio=False)
    np.testing.assert_((np.array(out.shape[:2]) - np.array(image.shape[:2]) + 1).all())

def test_image_random_flip():
    image = mx.random.normal(shape=(240, 120, 3)).astype(np.uint8)
    # no flip
    out, f = transforms.image.random_flip(image, 0, 0)
    np.testing.assert_allclose(image.asnumpy(), out.asnumpy())
    assert(f == (False, False))
    # horizontal
    out, f = transforms.image.random_flip(image, 1, 0)
    np.testing.assert_allclose(image.asnumpy()[:, ::-1, :], out.asnumpy())
    assert(f == (True, False))
    # vertical
    out, f = transforms.image.random_flip(image, 0, 1)
    np.testing.assert_allclose(image.asnumpy()[::-1, :, :], out.asnumpy())
    assert(f == (False, True))
    # both
    out, f = transforms.image.random_flip(image, 1, 1)
    np.testing.assert_allclose(image.asnumpy()[::-1, ::-1, :], out.asnumpy())
    assert(f == (True, True))

def test_image_resize_contain():
    image = mx.random.normal(shape=(240, 120, 3)).astype(np.uint8)
    width = 123
    height = 321
    out, _ = transforms.image.resize_contain(image, (width, height))
    np.testing.assert_allclose(out.shape, (height, width, 3))

    width = 120
    height = 20
    out, _ = transforms.image.resize_contain(image, (width, height))
    np.testing.assert_allclose(out.shape, (height, width, 3))

def test_image_ten_crop():
    image = mx.random.normal(shape=(240, 120, 3)).astype(np.uint8)
    size = (24, 24)
    crops = transforms.image.ten_crop(image, size)
    assert len(crops) == 10
    im = image.asnumpy()[:24, :24, :]
    np.testing.assert_allclose(im, crops[1].asnumpy())
    np.testing.assert_allclose(im[:, ::-1, :], crops[6].asnumpy())

def test_experimental_bbox_random_crop_with_constraints():
    bbox = np.array([[10, 20, 200, 500], [150, 200, 400, 300]])
    size = (640, 480)
    for _ in range(10):
        min_scale = np.random.uniform(0, 0.9)
        max_scale = np.random.uniform(min_scale, 1)
        max_aspect_ratio = np.random.uniform(1, 3)
        out, crop = transforms.experimental.bbox.random_crop_with_constraints(
            bbox, size, min_scale=min_scale, max_scale=max_scale,
            max_aspect_ratio=max_aspect_ratio, max_trial=20)
    assert out.size >= 4

def test_experimental_image_random_color_distort():
    image = mx.random.normal(shape=(240, 120, 3)).astype(np.float32)
    for _ in range(10):
        brightness_delta = np.random.randint(0, 64)
        contrast_low = np.random.uniform(0, 1)
        contrast_high = np.random.uniform(1, 2)
        saturation_low = np.random.uniform(0, 1)
        saturation_high = np.random.uniform(1, 2)
        hue_delta = np.random.randint(0, 36)
        out = transforms.experimental.image.random_color_distort(
            image, brightness_delta=brightness_delta, contrast_low=contrast_low,
            contrast_high=contrast_high, saturation_low=saturation_low,
            saturation_high=saturation_high, hue_delta=hue_delta)
        np.testing.assert_allclose(out.shape, image.shape)

def test_transforms_presets_ssd():
    im_fname = gcv.utils.download('https://github.com/dmlc/web-data/blob/master/' +
                                  'gluoncv/detection/biking.jpg?raw=true', path='biking.jpg')
    x, orig_img = ssd.load_test(im_fname, short=512)
    x1, orig_img1 = ssd.transform_test(mx.image.imread(im_fname), short=512)
    np.testing.assert_allclose(x.asnumpy(), x1.asnumpy())
    np.testing.assert_allclose(orig_img, orig_img1)
    if not osp.isdir(osp.expanduser('~/.mxnet/datasets/voc')):
        return
    train_dataset = VOCDetectionTiny()
    val_dataset = VOCDetectionTiny(splits=[('tiny_motorbike', 'test')])
    width, height = (512, 512)
    net = gcv.model_zoo.get_model('ssd_512_resnet50_v1_voc', pretrained=False, pretrained_base=False)
    net.initialize()
    num_workers = 0
    batch_size = 4
    with autograd.train_mode():
        _, _, anchors = net(mx.nd.zeros((1, 3, height, width)))
    batchify_fn = Tuple(Stack(), Stack(), Stack())  # stack image, cls_targets, box_targets
    train_loader = gluon.data.DataLoader(
        train_dataset.transform(ssd.SSDDefaultTrainTransform(width, height, anchors)),
        batch_size, True, batchify_fn=batchify_fn, last_batch='rollover', num_workers=num_workers)
    val_batchify_fn = Tuple(Stack(), Pad(pad_val=-1))
    val_loader = gluon.data.DataLoader(
        val_dataset.transform(ssd.SSDDefaultValTransform(width, height)),
        batch_size, False, batchify_fn=val_batchify_fn, last_batch='keep', num_workers=num_workers)
    train_loader2 = gluon.data.DataLoader(
        train_dataset.transform(ssd.SSDDefaultTrainTransform(width, height)),
        batch_size, True, batchify_fn=val_batchify_fn, last_batch='rollover', num_workers=num_workers)

    for loader in [train_loader, val_loader, train_loader2]:
        for i, batch in enumerate(loader):
            if i > 1:
                break
            pass

def test_transforms_presets_rcnn():
    im_fname = gcv.utils.download('https://github.com/dmlc/web-data/blob/master/' +
                                  'gluoncv/detection/biking.jpg?raw=true', path='biking.jpg')
    x, orig_img = rcnn.load_test(im_fname, short=600, max_size=1000)
    x1, orig_img1 = rcnn.transform_test(mx.image.imread(im_fname), short=600, max_size=1000)
    np.testing.assert_allclose(x.asnumpy(), x1.asnumpy())
    np.testing.assert_allclose(orig_img, orig_img1)
    if not osp.isdir(osp.expanduser('~/.mxnet/datasets/voc')):
        return
    train_dataset = VOCDetectionTiny()
    val_dataset = VOCDetectionTiny(splits=[('tiny_motorbike', 'test')])
    width, height = (512, 512)
    net = gcv.model_zoo.get_model('faster_rcnn_resnet50_v1b_voc', pretrained=False, pretrained_base=False)
    net.initialize()
    num_workers = 0
    short, max_size = 600, 1000
    batch_size = 4
    train_bfn = batchify.Tuple(*[batchify.Append() for _ in range(5)])
    train_loader = mx.gluon.data.DataLoader(
        train_dataset.transform(rcnn.FasterRCNNDefaultTrainTransform(short, max_size, net, flip_p=0.5)),
        batch_size, True, batchify_fn=train_bfn, last_batch='rollover', num_workers=num_workers)
    val_bfn = batchify.Tuple(*[batchify.Append() for _ in range(3)])
    val_loader = mx.gluon.data.DataLoader(
        val_dataset.transform(rcnn.FasterRCNNDefaultValTransform(short, max_size)),
        batch_size, False, batchify_fn=val_bfn, last_batch='keep', num_workers=num_workers)
    train_loader2 = gluon.data.DataLoader(
        train_dataset.transform(rcnn.FasterRCNNDefaultTrainTransform(short, max_size)),
        batch_size, True, batchify_fn=batchify.Tuple(*[batchify.Append() for _ in range(2)]),
        last_batch='rollover', num_workers=num_workers)

    for loader in [train_loader, val_loader, train_loader2]:
        for i, batch in enumerate(loader):
            if i > 1:
                break
            pass

def test_transforms_presets_mask_rcnn():
    # use valid only, loading training split is very slow
    train_dataset = COCOInstanceTiny(root=osp.join('~', '.mxnet', 'datasets', 'tiny_coco'),
                                          splits=('instances_val2017_tiny',), skip_empty=True)
    val_dataset = COCOInstanceTiny(root=osp.join('~', '.mxnet', 'datasets', 'tiny_coco'),
                                        splits=('instances_val2017_tiny',))
    net = gcv.model_zoo.get_model('mask_rcnn_resnet50_v1b_coco', pretrained=False, pretrained_base=False)
    net.initialize()
    num_workers = 0
    short, max_size = 800, 1333
    batch_size = 8
    train_bfn = batchify.Tuple(*[batchify.Append() for _ in range(6)])
    train_loader = mx.gluon.data.DataLoader(
        train_dataset.transform(rcnn.MaskRCNNDefaultTrainTransform(short, max_size, net)),
        batch_size, True, batchify_fn=train_bfn, last_batch='rollover', num_workers=num_workers)
    val_bfn = batchify.Tuple(*[batchify.Append() for _ in range(2)])
    val_loader = mx.gluon.data.DataLoader(
        val_dataset.transform(rcnn.MaskRCNNDefaultValTransform(short, max_size)),
        batch_size, False, batchify_fn=val_bfn, last_batch='keep', num_workers=num_workers)

    for loader in [train_loader, val_loader]:
        for i, batch in enumerate(loader):
            if i > 1:
                break
            pass

def test_transforms_presets_yolo():
    im_fname = gcv.utils.download('https://github.com/dmlc/web-data/blob/master/' +
                                  'gluoncv/detection/biking.jpg?raw=true', path='biking.jpg')
    x, orig_img = yolo.load_test(im_fname, short=512)
    x1, orig_img1 = yolo.transform_test(mx.image.imread(im_fname), short=512)
    np.testing.assert_allclose(x.asnumpy(), x1.asnumpy())
    np.testing.assert_allclose(orig_img, orig_img1)
    if not osp.isdir(osp.expanduser('~/.mxnet/datasets/voc')):
        return
    train_dataset = VOCDetectionTiny()
    val_dataset = VOCDetectionTiny(splits=[('tiny_motorbike', 'test')])
    width, height = (512, 512)
    net = gcv.model_zoo.get_model('yolo3_darknet53_voc', pretrained=False, pretrained_base=False)
    net.initialize()
    num_workers = 0
    batch_size = 4
    batchify_fn = Tuple(*([Stack() for _ in range(6)] + [Pad(axis=0, pad_val=-1) for _ in range(1)]))
    train_loader = gluon.data.DataLoader(
        train_dataset.transform(yolo.YOLO3DefaultTrainTransform(width, height, net)),
        batch_size, True, batchify_fn=batchify_fn, last_batch='rollover', num_workers=num_workers)
    val_batchify_fn = Tuple(Stack(), Pad(pad_val=-1))
    val_loader = gluon.data.DataLoader(
        val_dataset.transform(yolo.YOLO3DefaultValTransform(width, height)),
        batch_size, False, batchify_fn=val_batchify_fn, last_batch='keep', num_workers=num_workers)
    train_loader2 = gluon.data.DataLoader(
        train_dataset.transform(yolo.YOLO3DefaultTrainTransform(width, height)),
        batch_size, True, batchify_fn=val_batchify_fn, last_batch='rollover', num_workers=num_workers)

    for loader in [train_loader, val_loader, train_loader2]:
        for i, batch in enumerate(loader):
            if i > 1:
                break
            pass

def test_transforms_presets_center_net():
    im_fname = gcv.utils.download('https://github.com/dmlc/web-data/blob/master/' +
                                  'gluoncv/detection/biking.jpg?raw=true', path='biking.jpg')
    x, orig_img = center_net.load_test(im_fname, short=512)
    x1, orig_img1 = center_net.transform_test(mx.image.imread(im_fname), short=512)
    np.testing.assert_allclose(x.asnumpy(), x1.asnumpy())
    np.testing.assert_allclose(orig_img, orig_img1)
    if not osp.isdir(osp.expanduser('~/.mxnet/datasets/voc')):
        return
    train_dataset = VOCDetectionTiny()
    val_dataset = VOCDetectionTiny(splits=[('tiny_motorbike', 'test')])
    width, height = (512, 512)
    net = gcv.model_zoo.get_model('center_net_resnet18_v1b_voc', pretrained=False, pretrained_base=False)
    net.initialize()
    num_workers = 0
    batch_size = 4
    batchify_fn = Tuple([Stack() for _ in range(6)])
    train_loader = gluon.data.DataLoader(
        train_dataset.transform(center_net.CenterNetDefaultTrainTransform(width, height, num_class=len(train_dataset.classes), scale_factor=net.scale)),
        batch_size, True, batchify_fn=batchify_fn, last_batch='rollover', num_workers=num_workers)
    val_batchify_fn = Tuple(Stack(), Pad(pad_val=-1))
    val_loader = gluon.data.DataLoader(
        val_dataset.transform(center_net.CenterNetDefaultValTransform(width, height)),
        batch_size, False, batchify_fn=val_batchify_fn, last_batch='keep', num_workers=num_workers)

    for loader in [train_loader, val_loader]:
        for i, batch in enumerate(loader):
            if i > 1:
                break
            pass

if __name__ == '__main__':
    import nose
    nose.runmodule()



"""
Tractor module for SMOT
"""
# pylint: disable=arguments-differ
from abc import ABC, abstractmethod
import numpy as np

import mxnet as mx

from .utils import timeit
from .general_detector import GeneralDetector


class BaseAnchorBasedTracktor(ABC):
    """
    BaseAnchorBasedTracktor
    """

    @abstractmethod
    def anchors(self):
        """
        Returns the list of anchors used in this detector.
        -------

        """
        raise NotImplementedError

    @abstractmethod
    def prepare_for_frame(self, frame):
        """
        This method should run anything that needs to happen before the motion prediction.
        It can prepare the detector or even run the backbone feature extractions.
        It can also  provide data to motion prediction
        Parameters
        ----------
        frame: the frame data, the same as in the detect_and_track method
        Returns
        -------
            motion_predict_data: optional data provided to motion prediction, if no data is provided, return None
        """
        raise NotImplementedError

    @abstractmethod
    def detect_and_track(self, frame, tracking_anchor_indices, tracking_anchor_weights, tracking_classes):
        """
        Perform detection and tracking on the new frame

        Parameters
        ----------
        frame: HxWx3 RGB image
        tracking_anchor_indices: NxM ndarray
        tracking_anchor_weights NxM ndarray
        tracking_classes: Nx1 ndarray of the class ids of the tracked object

        Returns
            detection_bounding_boxes: all detection results, in (x0, y0, x1, y1, confidence, cls) format
            detection_source: source anchor box indices for each detection
            tracking_boxes: all tracking results, in (x0, y0, x1, y1, confidence) format
            extract_info: extra information from the tracktor, e.g. landmarks, a dict
        -------

        """
        raise NotImplementedError

    @abstractmethod
    def clean_up(self):
        """
        Clean up after running one video
        Returns
        -------

        """
        raise NotImplementedError



class GluonSSDMultiClassTracktor(BaseAnchorBasedTracktor):
    """
    Initiate a tracktor based on an object detetor.
    """
    def __init__(self, gpu_id=0, detector_thresh=0.5, model_name="",
                 use_pretrained=False, param_path="", data_shape=512):

        self.detector = GeneralDetector(gpu_id,
                                        aspect_ratio=1.,
                                        data_shape=data_shape,
                                        model_name=model_name,
                                        use_pretrained=use_pretrained,
                                        param_path=param_path
                                        )
        self._anchor_tensor = None
        self._detector_thresh = detector_thresh
        self._ctx = mx.gpu(gpu_id)

        self._dummy_ti = mx.nd.array([[0]], ctx=self._ctx)

    def anchors(self):
        if self.detector.anchor_tensor is None:
            raise ValueError("anchor not initialized yet")
        return self.detector.anchor_tensor

    def clean_up(self):
        pass

    def prepare_for_frame(self, frame):
        return None

    @timeit
    def detect_and_track(self, frame, tracking_anchor_indices, tracking_anchor_weights, tracking_object_classes):

        with_tracking = len(tracking_anchor_indices) > 0

        if with_tracking:
            tracking_indices = mx.nd.array(tracking_anchor_indices, ctx=self._ctx)
            tracking_weights = mx.nd.array(tracking_anchor_weights, ctx=self._ctx)
            tracking_classes = mx.nd.array(tracking_object_classes.reshape((-1, 1)), ctx=self._ctx)
        else:
            tracking_classes = tracking_indices = tracking_weights = self._dummy_ti

        ids, scores, bboxes, voted_tracking_bboxes, detection_anchor_indices = \
            self.detector.run_detection(frame, tracking_indices, tracking_weights, tracking_classes)

        valid_det_num = (scores > self._detector_thresh).sum().astype(int).asnumpy()[0]

        if valid_det_num > 0:
            valid_scores = scores[:valid_det_num]
            valid_bboxes = bboxes[:valid_det_num, :]
            valid_classes = ids[:valid_det_num, :]
            detection_output = mx.nd.concat(valid_bboxes, valid_scores, valid_classes, dim=1).asnumpy()
            anchor_indices_output = detection_anchor_indices[:valid_det_num, :]
        else:
            # no detection
            detection_output = np.array([])
            anchor_indices_output = np.array([])

        tracking_response = voted_tracking_bboxes.asnumpy() \
            if with_tracking else np.array([])

        return detection_output, anchor_indices_output, tracking_response, \
               {}



# adapted from https://github.com/facebookresearch/detectron2/blob/master/detectron2/data/build.py
# pylint: disable=bad-continuation,bad-whitespace,pointless-string-statement,redefined-outer-name,wrong-import-order
import itertools
import random
import logging
import operator
import dill
import pickle
from pprint import pformat
from typing import List, Optional, Union
import numpy as np
import torch
import torch.utils.data as data

from ..structures import BoxMode
from ...utils.comm import get_world_size
from ...utils.random import seed_all_rng
from ..transforms import instance_transforms as T

from ..registry.catalog import DatasetCatalog, MetadataCatalog
from .detection_utils import check_metadata_consistency
from . import detection_utils as utils
from .samplers import InferenceSampler, RepeatFactorTrainingSampler, TrainingSampler

"""
This file contains the default logic to build a dataloader for training or testing.
"""

__all__ = [
    "build_batch_data_loader",
    "build_detection_train_loader",
    "build_detection_test_loader",
    "get_detection_dataset_dicts",
    "load_proposals_into_dataset",
    "print_instances_class_histogram",
    "trivial_batch_collator",
]


def filter_images_with_only_crowd_annotations(dataset_dicts):
    """
    Filter out images with none annotations or only crowd annotations
    (i.e., images without non-crowd annotations).
    A common training-time preprocessing on COCO dataset.

    Args:
        dataset_dicts (list[dict]): annotations in Detectron2 Dataset format.

    Returns:
        list[dict]: the same format, but filtered.
    """
    num_before = len(dataset_dicts)

    def valid(anns):
        for ann in anns:
            if ann.get("iscrowd", 0) == 0:
                return True
        return False

    dataset_dicts = [x for x in dataset_dicts if valid(x["annotations"])]
    num_after = len(dataset_dicts)
    logger = logging.getLogger(__name__)
    logger.info(
        "Removed {} images with no usable annotations. {} images left.".format(
            num_before - num_after, num_after
        )
    )
    return dataset_dicts


def filter_images_with_few_keypoints(dataset_dicts, min_keypoints_per_image):
    """
    Filter out images with too few number of keypoints.

    Args:
        dataset_dicts (list[dict]): annotations in Detectron2 Dataset format.

    Returns:
        list[dict]: the same format as dataset_dicts, but filtered.
    """
    num_before = len(dataset_dicts)

    def visible_keypoints_in_image(dic):
        # Each keypoints field has the format [x1, y1, v1, ...], where v is visibility
        annotations = dic["annotations"]
        return sum(
            (np.array(ann["keypoints"][2::3]) > 0).sum()
            for ann in annotations
            if "keypoints" in ann
        )

    dataset_dicts = [
        x for x in dataset_dicts if visible_keypoints_in_image(x) >= min_keypoints_per_image
    ]
    num_after = len(dataset_dicts)
    logger = logging.getLogger(__name__)
    logger.info(
        "Removed {} images with fewer than {} keypoints.".format(
            num_before - num_after, min_keypoints_per_image
        )
    )
    return dataset_dicts


def load_proposals_into_dataset(dataset_dicts, proposal_file):
    """
    Load precomputed object proposals into the dataset.

    The proposal file should be a pickled dict with the following keys:

    - "ids": list[int] or list[str], the image ids
    - "boxes": list[np.ndarray], each is an Nx4 array of boxes corresponding to the image id
    - "objectness_logits": list[np.ndarray], each is an N sized array of objectness scores
      corresponding to the boxes.
    - "bbox_mode": the BoxMode of the boxes array. Defaults to ``BoxMode.XYXY_ABS``.

    Args:
        dataset_dicts (list[dict]): annotations in Detectron2 Dataset format.
        proposal_file (str): file path of pre-computed proposals, in pkl format.

    Returns:
        list[dict]: the same format as dataset_dicts, but added proposal field.
    """
    logger = logging.getLogger(__name__)
    logger.info("Loading proposals from: {}".format(proposal_file))

    with open(proposal_file, "rb") as f:
        proposals = pickle.load(f, encoding="latin1")

    # Rename the key names in D1 proposal files
    rename_keys = {"indexes": "ids", "scores": "objectness_logits"}
    for key in rename_keys:
        if key in proposals:
            proposals[rename_keys[key]] = proposals.pop(key)

    # Fetch the indexes of all proposals that are in the dataset
    # Convert image_id to str since they could be int.
    img_ids = set({str(record["image_id"]) for record in dataset_dicts})
    id_to_index = {str(id): i for i, id in enumerate(proposals["ids"]) if str(id) in img_ids}

    # Assuming default bbox_mode of precomputed proposals are 'XYXY_ABS'
    bbox_mode = BoxMode(proposals["bbox_mode"]) if "bbox_mode" in proposals else BoxMode.XYXY_ABS

    for record in dataset_dicts:
        # Get the index of the proposal
        i = id_to_index[str(record["image_id"])]

        boxes = proposals["boxes"][i]
        objectness_logits = proposals["objectness_logits"][i]
        # Sort the proposals in descending order of the scores
        inds = objectness_logits.argsort()[::-1]
        record["proposal_boxes"] = boxes[inds]
        record["proposal_objectness_logits"] = objectness_logits[inds]
        record["proposal_bbox_mode"] = bbox_mode

    return dataset_dicts


def print_instances_class_histogram(dataset_dicts, class_names):
    """
    Args:
        dataset_dicts (list[dict]): list of dataset dicts.
        class_names (list[str]): list of class names (zero-indexed).
    """
    num_classes = len(class_names)
    hist_bins = np.arange(num_classes + 1)
    histogram = np.zeros((num_classes,), dtype=np.int)
    for entry in dataset_dicts:
        annos = entry["annotations"]
        classes = [x["category_id"] for x in annos if not x.get("iscrowd", 0)]
        histogram += np.histogram(classes, bins=hist_bins)[0]

    N_COLS = min(6, len(class_names) * 2)

    def short_name(x):
        # make long class names shorter. useful for lvis
        if len(x) > 13:
            return x[:11] + ".."
        return x

    data = list(
        itertools.chain(*[[short_name(class_names[i]), int(v)] for i, v in enumerate(histogram)])
    )
    total_num_instances = sum(data[1::2])
    data.extend([None] * (N_COLS - (len(data) % N_COLS)))
    if num_classes > 1:
        data.extend(["total", total_num_instances])
    data = itertools.zip_longest(*[data[i::N_COLS] for i in range(N_COLS)])
    logging.info("Distribution of instances among all {} categories:\n".format(num_classes))
    logging.info(pformat(data))


def get_detection_dataset_dicts(
    dataset_names, filter_empty=True, min_keypoints=0, proposal_files=None
):
    """
    Load and prepare dataset dicts for instance detection/segmentation and semantic segmentation.

    Args:
        dataset_names (list[str]): a list of dataset names
        filter_empty (bool): whether to filter out images without instance annotations
        min_keypoints (int): filter out images with fewer keypoints than
            `min_keypoints`. Set to 0 to do nothing.
        proposal_files (list[str]): if given, a list of object proposal files
            that match each dataset in `dataset_names`.

    Returns:
        list[dict]: a list of dicts following the standard dataset dict format.
    """
    assert len(dataset_names)
    dataset_dicts = [DatasetCatalog.get(dataset_name) for dataset_name in dataset_names]
    for dataset_name, dicts in zip(dataset_names, dataset_dicts):
        assert len(dicts), "Dataset '{}' is empty!".format(dataset_name)

    if proposal_files is not None:
        assert len(dataset_names) == len(proposal_files)
        # load precomputed proposals from proposal files
        dataset_dicts = [
            load_proposals_into_dataset(dataset_i_dicts, proposal_file)
            for dataset_i_dicts, proposal_file in zip(dataset_dicts, proposal_files)
        ]

    dataset_dicts = list(itertools.chain.from_iterable(dataset_dicts))

    has_instances = "annotations" in dataset_dicts[0]
    if filter_empty and has_instances:
        dataset_dicts = filter_images_with_only_crowd_annotations(dataset_dicts)
    if min_keypoints > 0 and has_instances:
        dataset_dicts = filter_images_with_few_keypoints(dataset_dicts, min_keypoints)

    if has_instances:
        try:
            class_names = MetadataCatalog.get(dataset_names[0]).thing_classes
            check_metadata_consistency("thing_classes", dataset_names)
            print_instances_class_histogram(dataset_dicts, class_names)
        except AttributeError:  # class names are not available for this dataset
            pass
    return dataset_dicts


def build_batch_data_loader(
    dataset, sampler, total_batch_size, *, aspect_ratio_grouping=False, num_workers=0
):
    """
    Build a batched dataloader for training.

    Args:
        dataset (torch.utils.data.Dataset): map-style PyTorch dataset. Can be indexed.
        sampler (torch.utils.data.sampler.Sampler): a sampler that produces indices
        total_batch_size (int): total batch size across GPUs.
        aspect_ratio_grouping (bool): whether to group images with similar
            aspect ratio for efficiency. When enabled, it requires each
            element in dataset be a dict with keys "width" and "height".
        num_workers (int): number of parallel data loading workers

    Returns:
        iterable[list]. Length of each list is the batch size of the current
            GPU. Each element in the list comes from the dataset.
    """
    world_size = get_world_size()
    assert (
        total_batch_size > 0 and total_batch_size % world_size == 0
    ), "Total batch size ({}) must be divisible by the number of gpus ({}).".format(
        total_batch_size, world_size
    )

    batch_size = total_batch_size // world_size
    if aspect_ratio_grouping:
        data_loader = torch.utils.data.DataLoader(
            dataset,
            sampler=sampler,
            num_workers=num_workers,
            batch_sampler=None,
            collate_fn=operator.itemgetter(0),  # don't batch, but yield individual elements
            worker_init_fn=worker_init_reset_seed,
        )  # yield individual mapped dict
        return AspectRatioGroupedDataset(data_loader, batch_size)
    else:
        batch_sampler = torch.utils.data.sampler.BatchSampler(
            sampler, batch_size, drop_last=True
        )  # drop_last so the batch always have the same size
        return torch.utils.data.DataLoader(
            dataset,
            num_workers=num_workers,
            batch_sampler=batch_sampler,
            collate_fn=trivial_batch_collator,
            worker_init_fn=worker_init_reset_seed,
        )


def build_detection_train_loader(cfg, mapper=None):
    """
    A data loader is created by the following steps:

    1. Use the dataset names in config to query :class:`DatasetCatalog`, and obtain a list of dicts.
    2. Coordinate a random shuffle order shared among all processes (all GPUs)
    3. Each process spawn another few workers to process the dicts. Each worker will:
       * Map each metadata dict into another format to be consumed by the model.
       * Batch them by simply putting dicts into a list.

    The batched ``list[mapped_dict]`` is what this dataloader will yield.

    Args:
        cfg (CfgNode): the config
        mapper (callable): a callable which takes a sample (dict) from dataset and
            returns the format to be consumed by the model.
            By default it will be `DatasetMapper.from_config(cfg, True)`.

    Returns:
        an infinite iterator of training data
    """
    dataset_dicts = get_detection_dataset_dicts(
        cfg.CONFIG.DATA.DATASET.TRAIN,
        filter_empty=cfg.CONFIG.DATA.DETECTION.FILTER_EMPTY_ANNOTATIONS,
        min_keypoints=cfg.CONFIG.MODEL.ROI_KEYPOINT_HEAD.MIN_KEYPOINTS_PER_IMAGE
        if cfg.CONFIG.DATA.KEYPOINT_ON
        else 0,
        proposal_files=cfg.CONFIG.DATA.DATASET.PROPOSAL_FILES_TRAIN if cfg.CONFIG.DATA.LOAD_PROPOSALS else None,
    )
    dataset = DatasetFromList(dataset_dicts, copy=False)

    if mapper is None:
        mapper = DatasetMapper.from_config(cfg, True)
    dataset = MapDataset(dataset, mapper)

    sampler_name = cfg.CONFIG.DATA.DETECTION.SAMPLER_TRAIN
    logger = logging.getLogger(__name__)
    logger.info("Using training sampler {}".format(sampler_name))
    # TODO avoid if-else?
    if sampler_name == "TrainingSampler":
        sampler = TrainingSampler(len(dataset))
    elif sampler_name == "RepeatFactorTrainingSampler":
        repeat_factors = RepeatFactorTrainingSampler.repeat_factors_from_category_frequency(
            dataset_dicts, cfg.CONFIG.DATA.DETECTION.REPEAT_THRESHOLD
        )
        sampler = RepeatFactorTrainingSampler(repeat_factors)
    else:
        raise ValueError("Unknown training sampler: {}".format(sampler_name))
    return build_batch_data_loader(
        dataset,
        sampler,
        cfg.CONFIG.TRAIN.BATCH_SIZE,
        aspect_ratio_grouping=cfg.CONFIG.DATA.DETECTION.ASPECT_RATIO_GROUPING,
        num_workers=cfg.CONFIG.DATA.NUM_WORKERS,
    )


def build_detection_test_loader(cfg, dataset_name, mapper=None):
    """
    Similar to `build_detection_train_loader`.
    But this function uses the given `dataset_name` argument (instead of the names in cfg),
    and uses batch size 1.

    Args:
        cfg: a detectron2 CfgNode
        dataset_name (str): a name of the dataset that's available in the DatasetCatalog
        mapper (callable): a callable which takes a sample (dict) from dataset
           and returns the format to be consumed by the model.
           By default it will be `DatasetMapper.from_config(cfg, False)`.

    Returns:
        DataLoader: a torch DataLoader, that loads the given detection
        dataset, with test-time transformation and batching.
    """
    dataset_dicts = get_detection_dataset_dicts(
        [dataset_name],
        filter_empty=False,
        proposal_files=[
            cfg.CONFIG.DATA.DATASET.PROPOSAL_FILES_TEST[list(cfg.CONFIG.DATA.DATASET.PROPOSAL_FILES_TEST).index(dataset_name)]
        ]
        if cfg.CONFIG.DATA.LOAD_PROPOSALS
        else None,
    )

    dataset = DatasetFromList(dataset_dicts)
    if mapper is None:
        mapper = DatasetMapper.from_config(cfg, False)
    dataset = MapDataset(dataset, mapper)

    sampler = InferenceSampler(len(dataset))
    # Always use 1 image per worker during inference since this is the
    # standard when reporting inference time in papers.
    batch_sampler = torch.utils.data.sampler.BatchSampler(sampler, 1, drop_last=False)

    data_loader = torch.utils.data.DataLoader(
        dataset,
        num_workers=cfg.CONFIG.DATA.NUM_WORKERS,
        batch_sampler=batch_sampler,
        collate_fn=trivial_batch_collator,
    )
    return data_loader


def trivial_batch_collator(batch):
    """
    A batch collator that does nothing.
    """
    return batch


def worker_init_reset_seed(worker_id):
    seed_all_rng(np.random.randint(2 ** 31) + worker_id)


class MapDataset(data.Dataset):
    """
    Map a function over the elements in a dataset.

    Args:
        dataset: a dataset where map function is applied.
        map_func: a callable which maps the element in dataset. map_func is
            responsible for error handling, when error happens, it needs to
            return None so the MapDataset will randomly use other
            elements from the dataset.
    """

    def __init__(self, dataset, map_func):
        self._dataset = dataset
        self._map_func = dill.dumps(map_func)

        self._rng = random.Random(42)
        self._fallback_candidates = set(range(len(dataset)))

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, idx):
        retry_count = 0
        cur_idx = int(idx)
        _map_func = dill.loads(self._map_func)

        while True:
            data = _map_func(self._dataset[cur_idx])
            if data is not None:
                self._fallback_candidates.add(cur_idx)
                return data

            # _map_func fails for this idx, use a random new index from the pool
            retry_count += 1
            self._fallback_candidates.discard(cur_idx)
            cur_idx = self._rng.sample(self._fallback_candidates, k=1)[0]

            if retry_count >= 3:
                logger = logging.getLogger(__name__)
                logger.warning(
                    "Failed to apply `_map_func` for idx: {}, retry count: {}".format(
                        idx, retry_count
                    )
                )


class DatasetFromList(data.Dataset):
    """
    Wrap a list to a torch Dataset. It produces elements of the list as data.
    """

    def __init__(self, lst: list, copy: bool = True, serialize: bool = True):
        """
        Args:
            lst (list): a list which contains elements to produce.
            copy (bool): whether to deepcopy the element when producing it,
                so that the result can be modified in place without affecting the
                source in the list.
            serialize (bool): whether to hold memory using serialized objects, when
                enabled, data loader workers can use shared RAM from master
                process instead of making a copy.
        """
        self._lst = lst
        self._copy = copy
        self._serialize = serialize

        def _serialize(data):
            buffer = pickle.dumps(data, protocol=-1)
            return np.frombuffer(buffer, dtype=np.uint8)

        if self._serialize:
            logger = logging.getLogger(__name__)
            logger.info(
                "Serializing {} elements to byte tensors and concatenating them all ...".format(
                    len(self._lst)
                )
            )
            self._lst = [_serialize(x) for x in self._lst]
            self._addr = np.asarray([len(x) for x in self._lst], dtype=np.int64)
            self._addr = np.cumsum(self._addr)
            self._lst = np.concatenate(self._lst)
            logger.info("Serialized dataset takes {:.2f} MiB".format(len(self._lst) / 1024 ** 2))

    def __len__(self):
        if self._serialize:
            return len(self._addr)
        else:
            return len(self._lst)

    def __getitem__(self, idx):
        if self._serialize:
            start_addr = 0 if idx == 0 else self._addr[idx - 1].item()
            end_addr = self._addr[idx].item()
            bytes = memoryview(self._lst[start_addr:end_addr])
            return pickle.loads(bytes)
        elif self._copy:
            return copy.deepcopy(self._lst[idx])
        else:
            return self._lst[idx]


class AspectRatioGroupedDataset(data.IterableDataset):
    """
    Batch data that have similar aspect ratio together.
    In this implementation, images whose aspect ratio < (or >) 1 will
    be batched together.
    This improves training speed because the images then need less padding
    to form a batch.

    It assumes the underlying dataset produces dicts with "width" and "height" keys.
    It will then produce a list of original dicts with length = batch_size,
    all with similar aspect ratios.
    """

    def __init__(self, dataset, batch_size):
        """
        Args:
            dataset: an iterable. Each element must be a dict with keys
                "width" and "height", which will be used to batch data.
            batch_size (int):
        """
        self.dataset = dataset
        self.batch_size = batch_size
        self._buckets = [[] for _ in range(2)]
        # Hard-coded two aspect ratio groups: w > h and w < h.
        # Can add support for more aspect ratio groups, but doesn't seem useful

    def __iter__(self):
        for d in self.dataset:
            w, h = d["width"], d["height"]
            bucket_id = 0 if w > h else 1
            bucket = self._buckets[bucket_id]
            bucket.append(d)
            if len(bucket) == self.batch_size:
                yield bucket[:]
                del bucket[:]


class DatasetMapper:
    """
    A callable which takes a dataset dict in Detectron2 Dataset format,
    and map it into a format used by the model.

    This is the default callable to be used to map your dataset dict into training data.
    You may need to follow it to implement your own one for customized logic,
    such as a different way to read or transform images.
    See :doc:`/tutorials/data_loading` for details.

    The callable currently does the following:

    1. Read the image from "file_name"
    2. Applies cropping/geometric transforms to the image and annotations
    3. Prepare data and annotations to Tensor and :class:`Instances`
    """

    def __init__(
        self,
        is_train: bool,
        *,
        augmentations: List[Union[T.Augmentation, T.Transform]],
        image_format: str,
        use_instance_mask: bool = False,
        use_keypoint: bool = False,
        instance_mask_format: str = "polygon",
        keypoint_hflip_indices: Optional[np.ndarray] = None,
        precomputed_proposal_topk: Optional[int] = None,
        recompute_boxes: bool = False
    ):
        """
        NOTE: this interface is experimental.

        Args:
            is_train: whether it's used in training or inference
            augmentations: a list of augmentations or deterministic transforms to apply
            image_format: an image format supported by :func:`detection_utils.read_image`.
            use_instance_mask: whether to process instance segmentation annotations, if available
            use_keypoint: whether to process keypoint annotations if available
            instance_mask_format: one of "polygon" or "bitmask". Process instance segmentation
                masks into this format.
            keypoint_hflip_indices: see :func:`detection_utils.create_keypoint_hflip_indices`
            precomputed_proposal_topk: if given, will load pre-computed
                proposals from dataset_dict and keep the top k proposals for each image.
            recompute_boxes: whether to overwrite bounding box annotations
                by computing tight bounding boxes from instance mask annotations.
        """
        if recompute_boxes:
            assert use_instance_mask, "recompute_boxes requires instance masks"
        # fmt: off
        self.is_train               = is_train
        self.augmentations          = augmentations
        self.image_format           = image_format
        self.use_instance_mask      = use_instance_mask
        self.instance_mask_format   = instance_mask_format
        self.use_keypoint           = use_keypoint
        self.keypoint_hflip_indices = keypoint_hflip_indices
        self.proposal_topk          = precomputed_proposal_topk
        self.recompute_boxes        = recompute_boxes
        # fmt: on
        logger = logging.getLogger(__name__)
        logger.info("Augmentations used in training: " + str(augmentations))

    @classmethod
    def from_config(cls, cfg, is_train: bool = True):
        augs = utils.build_augmentation(cfg, is_train)
        if cfg.CONFIG.DATA.CROP.ENABLED and is_train:
            augs.insert(0, T.RandomCrop(cfg.CONFIG.DATA.CROP.TYPE, cfg.CONFIG.DATA.CROP.SIZE))
            recompute_boxes = cfg.CONFIG.DATA.MASK_ON
        else:
            recompute_boxes = False

        ret = {
            "is_train": is_train,
            "augmentations": augs,
            "image_format": "RGB",
            "use_instance_mask": cfg.CONFIG.DATA.MASK_ON,
            "instance_mask_format": cfg.CONFIG.DATA.MASK_FORMAT,
            "use_keypoint": cfg.CONFIG.DATA.KEYPOINT_ON,
            "recompute_boxes": recompute_boxes,
        }
        if cfg.CONFIG.DATA.KEYPOINT_ON:
            ret["keypoint_hflip_indices"] = utils.create_keypoint_hflip_indices(cfg.CONFIG.DATA.DATASET.TRAIN)

        if cfg.CONFIG.DATA.LOAD_PROPOSALS:
            ret["precomputed_proposal_topk"] = (
                cfg.CONFIG.DATA.DATASET.PRECOMPUTED_PROPOSAL_TOPK_TRAIN
                if is_train
                else cfg.CONFIG.DATA.DATASET.PRECOMPUTED_PROPOSAL_TOPK_TEST
            )
        return ret

    def __call__(self, dataset_dict):
        """
        Args:
            dataset_dict (dict): Metadata of one image, in Detectron2 Dataset format.

        Returns:
            dict: a format that builtin models in detectron2 accept
        """
        dataset_dict = copy.deepcopy(dataset_dict)  # it will be modified by code below
        # USER: Write your own image loading if it's not from a file
        image = utils.read_image(dataset_dict["file_name"], format=self.image_format)
        utils.check_image_size(dataset_dict, image)

        # USER: Remove if you don't do semantic/panoptic segmentation.
        if "sem_seg_file_name" in dataset_dict:
            sem_seg_gt = utils.read_image(dataset_dict.pop("sem_seg_file_name"), "L").squeeze(2)
        else:
            sem_seg_gt = None

        aug_input = T.StandardAugInput(image, sem_seg=sem_seg_gt)
        transforms = aug_input.apply_augmentations(self.augmentations)
        image, sem_seg_gt = aug_input.image, aug_input.sem_seg

        image_shape = image.shape[:2]  # h, w
        # Pytorch's dataloader is efficient on torch.Tensor due to shared-memory,
        # but not efficient on large generic data structures due to the use of pickle & mp.Queue.
        # Therefore it's important to use torch.Tensor.
        dataset_dict["image"] = torch.as_tensor(np.ascontiguousarray(image.transpose(2, 0, 1)))
        if sem_seg_gt is not None:
            dataset_dict["sem_seg"] = torch.as_tensor(sem_seg_gt.astype("long"))

        # USER: Remove if you don't use pre-computed proposals.
        # Most users would not need this feature.
        if self.proposal_topk is not None:
            utils.transform_proposals(
                dataset_dict, image_shape, transforms, proposal_topk=self.proposal_topk
            )

        if not self.is_train:
            # USER: Modify this if you want to keep them for some reason.
            dataset_dict.pop("annotations", None)
            dataset_dict.pop("sem_seg_file_name", None)
            return dataset_dict

        if "annotations" in dataset_dict:
            # USER: Modify this if you want to keep them for some reason.
            for anno in dataset_dict["annotations"]:
                if not self.use_instance_mask:
                    anno.pop("segmentation", None)
                if not self.use_keypoint:
                    anno.pop("keypoints", None)

            # USER: Implement additional transformations if you have other types of data
            annos = [
                utils.transform_instance_annotations(
                    obj, transforms, image_shape, keypoint_hflip_indices=self.keypoint_hflip_indices
                )
                for obj in dataset_dict.pop("annotations")
                if obj.get("iscrowd", 0) == 0
            ]
            instances = utils.annotations_to_instances(
                annos, image_shape, mask_format=self.instance_mask_format
            )

            # After transforms such as cropping are applied, the bounding box may no longer
            # tightly bound the object. As an example, imagine a triangle object
            # [(0,0), (2,0), (0,2)] cropped by a box [(1,0),(2,2)] (XYXY format). The tight
            # bounding box of the cropped triangle should be [(1,0),(2,1)], which is not equal to
            # the intersection of original bounding box and the cropping box.
            if self.recompute_boxes:
                instances.gt_boxes = instances.gt_masks.get_bounding_boxes()
            dataset_dict["instances"] = utils.filter_empty_instances(instances)
        return dataset_dict



"""Auto data preparation."""
# pylint: disable=bare-except
import os
import re
from pathlib import Path
import yaml
from ...utils.download import download
from ...utils.filesystem import unzip, untar, PathTree

__all__ = ['url_data']

def url_data(url, path=None, overwrite=False, overwrite_folder=False, sha1_hash=None, root=None, disp_depth=1):
    """Download dataset with a given URL

    Parameters
    ----------
    url : str
        URL to download
    path : str, optional
        Destination path to store downloaded file. By default stores to the
        ~/.gluoncv directory with same name as in url.
        You can also change the default behavior by editing ~/.gluoncv/config.yaml.
    overwrite : bool, optional
        Whether to overwrite destination file if already exists.
    overwrite_folder : bool, optional
        Whether to extract file to destination folder if already exists. You may use this option if you suspect
        the destination is corrupted or some files are missing.
    sha1_hash : str, optional
        Expected sha1 hash in hexadecimal digits. Will ignore existing file when hash is specified
        but doesn't match.
    root : str, optional
        Destination root dir to store extracted files. By default it's default in ~/.gluoncv directory.
    disp_depth : int, optional
        If set > 0, will print out the tree structure of extracted dataset folder with maximum `disp_depth`.

    Returns
    -------
    str or tuple of str
        The file path of the downloaded file.
    """
    fname = Path(path or URLs.path(url, c_key='archive'))
    fname.parent.mkdir(parents=True, exist_ok=True)
    fname = download(url, path=str(fname.resolve()), overwrite=overwrite, sha1_hash=sha1_hash)
    extract_root = URLs.path(url, c_key='data')
    extract_root = extract_root.parent.joinpath(extract_root.stem)
    extract_root.mkdir(parents=True, exist_ok=True)
    if fname.endswith('.zip'):
        folder = unzip(fname, root=root if root else extract_root, strict=overwrite_folder)
    elif fname.endswith('gz'):
        folder = untar(fname, root=root if root else extract_root, strict=overwrite_folder)
    else:
        raise ValueError('Unknown url data with file: {}'.format(fname))

    if disp_depth > 0:
        try:
            path_tree = PathTree(folder, disp_depth)
            print(path_tree)
        except:
            pass

    return Path(folder)


class Config:
    "Setup config at `~/.gluoncv` unless it exists already."
    config_path = Path(os.getenv('MXNET_HOME', '~/.gluoncv')).expanduser()
    config_file = config_path/'config.yml'

    def __init__(self):
        self.config_path.mkdir(parents=True, exist_ok=True)
        if not self.config_file.exists():
            self.create_config()
        self.d = self.load_config()

    def __getitem__(self, k):
        k = k.lower()
        if k not in self.d:
            k = k + '_path'
        return Path(self.d[k])

    def __getattr__(self, k):
        if k == 'd':
            raise AttributeError
        return self[k]

    def __setitem__(self, k, v):
        self.d[k] = str(v)
    def __contains__(self, k):
        return k in self.d

    def load_config(self):
        "load and return config if version equals 2 in existing, else create new config."
        with open(self.config_file, 'r') as f:
            config = yaml.safe_load(f)
            if 'version' in config and config['version'] == 2:
                return config
            elif 'version' in config:
                self.create_config(config)
            else:
                self.create_config()
        return self.load_config()

    def create_config(self, cfg=None):
        "create new config with default paths and set `version` to 2."
        config = {'data_path':    str(self.config_path/'datasets'),
                  'archive_path': str(self.config_path/'archive'),
                  'storage_path': '/tmp',
                  'model_path':   str(self.config_path/'models'),
                  'version':      2}
        if cfg is not None:
            cfg['version'] = 2
            config = merge(config, cfg)
        self.save_file(config)

    def save(self):
        self.save_file(self.d)

    def save_file(self, config):
        "save config file at default config location `~/.gluoncv/config.yml`."
        with self.config_file.open('w') as f:
            yaml.dump(config, f, default_flow_style=False)


# pylint: disable=bad-whitespace
class URLs():
    "Global constants for dataset and model URLs."
    LOCAL_PATH = Path.cwd()
    MDL = 'http://files.fast.ai/models/'
    S3  = 'https://s3.amazonaws.com/fast-ai-'
    URL = f'{S3}sample/'

    S3_IMAGE    = f'{S3}imageclas/'
    S3_IMAGELOC = f'{S3}imagelocal/'
    S3_COCO     = f'{S3}coco/'

    # main datasets
    ADULT_SAMPLE        = f'{URL}adult_sample.tgz'
    BIWI_SAMPLE         = f'{URL}biwi_sample.tgz'
    CIFAR               = f'{URL}cifar10.tgz'
    COCO_SAMPLE         = f'{S3_COCO}coco_sample.tgz'
    COCO_TINY           = f'{S3_COCO}coco_tiny.tgz'
    HUMAN_NUMBERS       = f'{URL}human_numbers.tgz'
    # IMDB                = f'{S3_NLP}imdb.tgz'
    IMDB_SAMPLE         = f'{URL}imdb_sample.tgz'
    ML_SAMPLE           = f'{URL}movie_lens_sample.tgz'
    ML_100k             = 'http://files.grouplens.org/datasets/movielens/ml-100k.zip'
    MNIST_SAMPLE        = f'{URL}mnist_sample.tgz'
    MNIST_TINY          = f'{URL}mnist_tiny.tgz'
    MNIST_VAR_SIZE_TINY = f'{S3_IMAGE}mnist_var_size_tiny.tgz'
    PLANET_SAMPLE       = f'{URL}planet_sample.tgz'
    PLANET_TINY         = f'{URL}planet_tiny.tgz'
    IMAGENETTE          = f'{S3_IMAGE}imagenette2.tgz'
    IMAGENETTE_160      = f'{S3_IMAGE}imagenette2-160.tgz'
    IMAGENETTE_320      = f'{S3_IMAGE}imagenette2-320.tgz'
    IMAGEWOOF           = f'{S3_IMAGE}imagewoof2.tgz'
    IMAGEWOOF_160       = f'{S3_IMAGE}imagewoof2-160.tgz'
    IMAGEWOOF_320       = f'{S3_IMAGE}imagewoof2-320.tgz'
    IMAGEWANG           = f'{S3_IMAGE}imagewang.tgz'
    IMAGEWANG_160       = f'{S3_IMAGE}imagewang-160.tgz'
    IMAGEWANG_320       = f'{S3_IMAGE}imagewang-320.tgz'

    # kaggle competitions download dogs-vs-cats -p {DOGS.absolute()}
    DOGS = f'{URL}dogscats.tgz'

    # image classification datasets
    CALTECH_101  = f'{S3_IMAGE}caltech_101.tgz'
    CARS         = f'{S3_IMAGE}stanford-cars.tgz'
    CIFAR_100    = f'{S3_IMAGE}cifar100.tgz'
    CUB_200_2011 = f'{S3_IMAGE}CUB_200_2011.tgz'
    FLOWERS      = f'{S3_IMAGE}oxford-102-flowers.tgz'
    FOOD         = f'{S3_IMAGE}food-101.tgz'
    MNIST        = f'{S3_IMAGE}mnist_png.tgz'
    PETS         = f'{S3_IMAGE}oxford-iiit-pet.tgz'

    # Image localization datasets
    BIWI_HEAD_POSE     = f"{S3_IMAGELOC}biwi_head_pose.tgz"
    CAMVID             = f'{S3_IMAGELOC}camvid.tgz'
    CAMVID_TINY        = f'{URL}camvid_tiny.tgz'
    LSUN_BEDROOMS      = f'{S3_IMAGE}bedroom.tgz'
    PASCAL_2007        = f'{S3_IMAGELOC}pascal_2007.tgz'
    PASCAL_2012        = f'{S3_IMAGELOC}pascal_2012.tgz'

    # Medical Imaging datasets
    #SKIN_LESION        = f'{S3_IMAGELOC}skin_lesion.tgz'
    SIIM_SMALL         = f'{S3_IMAGELOC}siim_small.tgz'

    @staticmethod
    def path(url='.', c_key='archive'):
        "Return local path where to download based on `c_key`"
        fname = url.split('/')[-1]
        local_path = URLs.LOCAL_PATH / ('models' if c_key == 'models' else 'datasets')/fname
        if local_path.exists():
            return local_path
        return Config()[c_key]/fname

_URL_REGEX = re.compile(
    r'^(?:http|ftp)s?://' # http:// or https://
    r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
    r'localhost|' #localhost...
    r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
    r'(?::\d+)?' # optional port
    r'(?:/?|[/?]\S+)$', re.IGNORECASE)

def is_url(url_like):
    if not isinstance(url_like, str):
        return False
    return re.match(_URL_REGEX, url_like) is not None



"""Default configs for torch image classification"""
# pylint: disable=bad-whitespace,missing-class-docstring
from typing import Union, Tuple
from autocfg import dataclass, field

@dataclass
class ImageClassification:
    model: str = 'resnet101'
    pretrained: bool = True
    global_pool_type: Union[str, None] = None  # Global pool type, one of (fast, avg, max, avgmax). Model default if None

@dataclass
class DataCfg:
    img_size: Union[int, None] = None  # Image patch size (default: None => model default)
    input_size: Union[Tuple[int, int, int], None] = None  # Input all image dimensions (d h w, e.g. --input-size 3 224 224), uses model default if empty
    crop_pct: Union[float, None] = 0.99  # Input image center crop percent (for validation only)
    mean: Union[Tuple, None] = None  # Override mean pixel value of dataset
    std : Union[Tuple, None] = None  # Override std deviation of of dataset
    interpolation: str = ''  # Image resize interpolation type (overrides model)
    validation_batch_size_multiplier: int = 1  # ratio of validation batch size to training batch size (default: 1)

@dataclass
class OptimizerCfg:
    opt: str = 'sgd'
    opt_eps: Union[float, None] = None  # Optimizer Epsilon (default: None, use opt default)
    opt_betas: Union[Tuple, None] = None  # Optimizer Betas (default: None, use opt default)
    momentum: float = 0.9
    weight_decay: float = 0.0001
    clip_grad: Union[float, None] = None  # Clip gradient norm (default: None, no clipping)
    clip_mode: str = 'norm'  # Gradient clipping mode. One of ("norm", "value", "agc")

@dataclass
class TrainCfg:
    batch_size: int = 32
    sched: str = 'step'  # LR scheduler
    lr: float = 0.01
    lr_noise: Union[Tuple, None] = None  # learning rate noise on/off epoch percentages
    lr_noise_pct: float = 0.67  # learning rate noise limit percent
    lr_noise_std: float = 1.0  # learning rate noise std-dev
    lr_cycle_mul: float = 1.0  # learning rate cycle len multiplier
    lr_cycle_limit: int = 1 # learning rate cycle limit
    transfer_lr_mult : float = 0.01  # reduce the backbone lr_mult to avoid quickly destroying the features
    output_lr_mult : float = 0.1  # the learning rate multiplier for last fc layer if trained with transfer learning
    warmup_lr: float = 0.0001
    min_lr: float = 1e-5
    epochs: int = 200
    start_epoch: int = 0  # manual epoch number (useful on restarts)
    decay_epochs: int = 30  # epoch interval to decay LR
    warmup_epochs: int = 3  # epochs to warmup LR, if scheduler supports
    cooldown_epochs: int = 10  # epochs to cooldown LR at min_lr, after cyclic schedule ends
    patience_epochs: int = 10  # patience epochs for Plateau LR scheduler
    decay_rate: float = 0.1
    bn_momentum: Union[float, None] = None  # BatchNorm momentum override
    bn_eps: Union[float, None] = None  # BatchNorm epsilon override
    sync_bn: bool = False  # Enable NVIDIA Apex or Torch synchronized BatchNorm
    early_stop_patience : int = -1  # epochs with no improvement after which train is early stopped, negative: disabled
    early_stop_min_delta : float = 0.001  # ignore changes less than min_delta for metrics
    # the baseline value for metric, training won't stop if not reaching baseline
    early_stop_baseline : Union[float, int] = 0.0
    early_stop_max_value : Union[float, int] = 1.0  # early stop if reaching max value instantly

@dataclass
class AugmentationCfg:
    no_aug: bool = False  # Disable all training augmentation, override other train aug args
    scale: Tuple[float, float] = (0.08, 1.0)  # Random resize scale
    ratio: Tuple[float, float] = (3./4., 4./3.) # Random resize aspect ratio (default: 0.75 1.33
    hflip: float = 0.5  # Horizontal flip training aug probability
    vflip: float = 0.0 # Vertical flip training aug probability
    color_jitter: float = 0.4
    auto_augment: Union[str, None] = None  # Use AutoAugment policy. "v0" or "original
    mixup: float = 0.0  # mixup alpha, mixup enabled if > 0
    cutmix: float = 0.0  # cutmix alpha, cutmix enabled if > 0
    cutmix_minmax: Union[Tuple, None] = None  # cutmix min/max ratio, overrides alpha and enables cutmix if set
    mixup_prob: float = 1.0  # Probability of performing mixup or cutmix when either/both is enabled
    mixup_switch_prob: float = 0.5  # Probability of switching to cutmix when both mixup and cutmix enabled
    mixup_mode: str = 'batch'  # How to apply mixup/cutmix params. Per "batch", "pair", or "elem"
    mixup_off_epoch: int = 0  # Turn off mixup after this epoch, disabled if 0
    smoothing: float = 0.1  # Label smoothin
    train_interpolation: str = 'random'  # Training interpolation (random, bilinear, bicubic)
    drop: float = 0.0  # Dropout rate
    drop_path: Union[float, None] = None  # Drop path rate
    drop_block: Union[float, None] = None # Drop block rate

@dataclass
class ModelEMACfg:
    model_ema: bool = True  # Enable tracking moving average of model weights
    model_ema_force_cpu: bool = False # Force ema to be tracked on CPU, rank=0 node only. Disables EMA validation
    model_ema_decay: float = 0.9998  # decay factor for model weights moving average

@dataclass
class MiscCfg:
    seed: int = 42
    log_interval: int = 50  # how many batches to wait before logging training status
    num_workers: int = 4  # how many training processes to use
    save_images: bool = False  # save images of input bathes every log interval for debugging
    amp: bool = False  # use NVIDIA Apex AMP or Native AMP for mixed precision training
    apex_amp: bool = False  # Use NVIDIA Apex AMP mixed precision
    native_amp: bool = False  # Use Native Torch AMP mixed precision
    pin_mem: bool = False  # Pin CPU memory in DataLoader for more efficient (sometimes) transfer to GPU
    prefetcher: bool = False  # use fast prefetcher
    eval_metric: str = 'top1'  # 'Best metric (default: "top1")
    tta: int = 0  # Test/inference time augmentation (oversampling) factor. 0=None
    use_multi_epochs_loader: bool = False  # use the multi-epochs-loader to save time at the beginning of every epoch
    torchscript: bool = False  # keep false, convert model torchscript for inference

@dataclass
class TorchImageClassificationCfg:
    img_cls : ImageClassification = field(default_factory=ImageClassification)
    data: DataCfg = field(default_factory=DataCfg)
    optimizer: OptimizerCfg = field(default_factory=OptimizerCfg)
    train: TrainCfg = field(default_factory=TrainCfg)
    augmentation: AugmentationCfg = field(default_factory=AugmentationCfg)
    model_ema: ModelEMACfg = field(default_factory=ModelEMACfg)
    misc: MiscCfg = field(default_factory=MiscCfg)
    gpus : Union[Tuple, list] = (0, )  # gpu individual ids, not necessarily consecutive



"""Group Norm implementation"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import Module, Parameter
from torch.nn import init

__all__ = ['GCN', 'NaiveGroupNorm']


class Conv2D(nn.Module):
    """Inline Conv2D for GroupNorm module"""
    def __init__(self, in_channels, out_channels, kernel_size, padding='same',
                 stride=1, dilation=1, groups=1):
        super(Conv2D, self).__init__()

        assert isinstance(kernel_size, (tuple, int)), "Allowed kernel type [int or tuple], not {}".format(type(kernel_size))
        assert padding == 'same', "Allowed padding type {}, not {}".format('same', padding)

        self.kernel_size = kernel_size
        if isinstance(kernel_size, tuple):
            self.h_kernel = kernel_size[0]
            self.w_kernel = kernel_size[1]
        else:
            self.h_kernel = kernel_size
            self.w_kernel = kernel_size

        self.padding = padding
        self.stride = stride
        self.dilation = dilation
        self.groups = groups
        self.conv = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size,
                              stride=self.stride, dilation=self.dilation, groups=self.groups)

    def forward(self, x):

        if self.padding == 'same':

            height, width = x.shape[2:]

            h_pad_need = max(0, (height - 1) * self.stride + self.h_kernel - height)
            w_pad_need = max(0, (width - 1) * self.stride + self.w_kernel - width)

            pad_left = w_pad_need // 2
            pad_right = w_pad_need - pad_left
            pad_top = h_pad_need // 2
            pad_bottom = h_pad_need - pad_top

            padding = (pad_left, pad_right, pad_top, pad_bottom)

            x = F.pad(x, padding, 'constant', 0)

        x = self.conv(x)

        return x


class GCN(nn.Module):
    """
    Large Kernel Matters -- https://arxiv.org/abs/1703.02719
    """
    def __init__(self, in_channels, out_channels, k=3):
        super(GCN, self).__init__()

        self.conv_l1 = Conv2D(in_channels=in_channels, out_channels=out_channels, kernel_size=(k, 1), padding='same')
        self.conv_l2 = Conv2D(in_channels=out_channels, out_channels=out_channels, kernel_size=(1, k), padding='same')

        self.conv_r1 = Conv2D(in_channels=in_channels, out_channels=out_channels, kernel_size=(1, k), padding='same')
        self.conv_r2 = Conv2D(in_channels=out_channels, out_channels=out_channels, kernel_size=(k, 1), padding='same')

    def forward(self, x):
        x1 = self.conv_l1(x)
        x1 = self.conv_l2(x1)

        x2 = self.conv_r1(x)
        x2 = self.conv_r2(x2)

        out = x1 + x2

        return out


class NaiveGroupNorm(Module):
    r"""NaiveGroupNorm implements Group Normalization with the high-level matrix operations in PyTorch.
    It is a temporary solution to export GN by ONNX before the official GN can be exported by ONNX.
    The usage of NaiveGroupNorm is exactly the same as the official :class:`torch.nn.GroupNorm`.
    Args:
        num_groups (int): number of groups to separate the channels into
        num_channels (int): number of channels expected in input
        eps: a value added to the denominator for numerical stability. Default: 1e-5
        affine: a boolean value that when set to ``True``, this module
            has learnable per-channel affine parameters initialized to ones (for weights)
            and zeros (for biases). Default: ``True``.

    Shape:
        - Input: :math:`(N, C, *)` where :math:`C=\text{num\_channels}`
        - Output: :math:`(N, C, *)` (same shape as input)

    Examples::

        >>> input = torch.randn(20, 6, 10, 10)
        >>> # Separate 6 channels into 3 groups
        >>> m = NaiveGroupNorm(3, 6)
        >>> # Separate 6 channels into 6 groups (equivalent with InstanceNorm)
        >>> m = NaiveGroupNorm(6, 6)
        >>> # Put all 6 channels into a single group (equivalent with LayerNorm)
        >>> m = NaiveGroupNorm(1, 6)
        >>> # Activating the module
        >>> output = m(input)

    .. _`Group Normalization`: https://arxiv.org/abs/1803.08494
    """
    __constants__ = ['num_groups', 'num_channels', 'eps', 'affine', 'weight',
                     'bias']

    def __init__(self, num_groups, num_channels, eps=1e-5, affine=True):
        super(NaiveGroupNorm, self).__init__()
        self.num_groups = num_groups
        self.num_channels = num_channels
        self.eps = eps
        self.affine = affine
        if self.affine:
            self.weight = Parameter(torch.Tensor(num_channels))
            self.bias = Parameter(torch.Tensor(num_channels))
        else:
            self.register_parameter('weight', None)
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        if self.affine:
            init.ones_(self.weight)
            init.zeros_(self.bias)

    def forward(self, input):
        N, C, H, W = input.size()
        assert C % self.num_groups == 0
        input = input.reshape(N, self.num_groups, -1)
        mean = input.mean(dim=-1, keepdim=True)
        var = (input ** 2).mean(dim=-1, keepdim=True) - mean ** 2
        std = torch.sqrt(var + self.eps)

        input = (input - mean) / std
        input = input.reshape(N, C, H, W)
        if self.affine:
            input = input * self.weight.reshape(1, C, 1, 1) + self.bias.reshape(1, C, 1, 1)
        return input

    def extra_repr(self):
        return '{num_groups}, {num_channels}, eps={eps}, ' \
            'affine={affine}'.format(**self.__dict__)



# adapted from https://github.com/facebookresearch/detectron2/blob/master/detectron2/data/samplers/distributed_sampler.py
import itertools
import math
from collections import defaultdict
from typing import Optional
import torch
from torch.utils.data.sampler import Sampler

from ....utils import comm


class TrainingSampler(Sampler):
    """
    In training, we only care about the "infinite stream" of training data.
    So this sampler produces an infinite stream of indices and
    all workers cooperate to correctly shuffle the indices and sample different indices.

    The samplers in each worker effectively produces `indices[worker_id::num_workers]`
    where `indices` is an infinite stream of indices consisting of
    `shuffle(range(size)) + shuffle(range(size)) + ...` (if shuffle is True)
    or `range(size) + range(size) + ...` (if shuffle is False)
    """

    def __init__(self, size: int, shuffle: bool = True, seed: Optional[int] = None):
        """
        Args:
            size (int): the total number of data of the underlying dataset to sample from
            shuffle (bool): whether to shuffle the indices or not
            seed (int): the initial seed of the shuffle. Must be the same
                across all workers. If None, will use a random seed shared
                among workers (require synchronization among all workers).
        """
        self._size = size
        assert size > 0
        self._shuffle = shuffle
        if seed is None:
            seed = comm.shared_random_seed()
        self._seed = int(seed)

        self._rank = comm.get_rank()
        self._world_size = comm.get_world_size()

    def __iter__(self):
        start = self._rank
        yield from itertools.islice(self._infinite_indices(), start, None, self._world_size)

    def _infinite_indices(self):
        g = torch.Generator()
        g.manual_seed(self._seed)
        while True:
            if self._shuffle:
                yield from torch.randperm(self._size, generator=g)
            else:
                yield from torch.arange(self._size)


class RepeatFactorTrainingSampler(Sampler):
    """
    Similar to TrainingSampler, but a sample may appear more times than others based
    on its "repeat factor". This is suitable for training on class imbalanced datasets like LVIS.
    """

    def __init__(self, repeat_factors, *, shuffle=True, seed=None):
        """
        Args:
            repeat_factors (Tensor): a float vector, the repeat factor for each indice. When it's
                full of ones, it is equivalent to ``TrainingSampler(len(repeat_factors), ...)``.
            shuffle (bool): whether to shuffle the indices or not
            seed (int): the initial seed of the shuffle. Must be the same
                across all workers. If None, will use a random seed shared
                among workers (require synchronization among all workers).
        """
        self._shuffle = shuffle
        if seed is None:
            seed = comm.shared_random_seed()
        self._seed = int(seed)

        self._rank = comm.get_rank()
        self._world_size = comm.get_world_size()

        # Split into whole number (_int_part) and fractional (_frac_part) parts.
        self._int_part = torch.trunc(repeat_factors)
        self._frac_part = repeat_factors - self._int_part

    @staticmethod
    def repeat_factors_from_category_frequency(dataset_dicts, repeat_thresh):
        """
        Compute (fractional) per-image repeat factors based on category frequency.
        The repeat factor for an image is a function of the frequency of the rarest
        category labeled in that image. The "frequency of category c" in [0, 1] is defined
        as the fraction of images in the training set (without repeats) in which category c
        appears.
        See :paper:`lvis` (>= v2) Appendix B.2.

        Args:
            dataset_dicts (list[dict]): annotations in Detectron2 dataset format.
            repeat_thresh (float): frequency threshold below which data is repeated.
                If the frequency is half of `repeat_thresh`, the image will be
                repeated twice.

        Returns:
            torch.Tensor: the i-th element is the repeat factor for the dataset image
                at index i.
        """
        # 1. For each category c, compute the fraction of images that contain it: f(c)
        category_freq = defaultdict(int)
        for dataset_dict in dataset_dicts:  # For each image (without repeats)
            cat_ids = {ann["category_id"] for ann in dataset_dict["annotations"]}
            for cat_id in cat_ids:
                category_freq[cat_id] += 1
        num_images = len(dataset_dicts)
        for k, v in category_freq.items():
            category_freq[k] = v / num_images

        # 2. For each category c, compute the category-level repeat factor:
        #    r(c) = max(1, sqrt(t / f(c)))
        category_rep = {
            cat_id: max(1.0, math.sqrt(repeat_thresh / cat_freq))
            for cat_id, cat_freq in category_freq.items()
        }

        # 3. For each image I, compute the image-level repeat factor:
        #    r(I) = max_{c in I} r(c)
        rep_factors = []
        for dataset_dict in dataset_dicts:
            cat_ids = {ann["category_id"] for ann in dataset_dict["annotations"]}
            rep_factor = max({category_rep[cat_id] for cat_id in cat_ids})
            rep_factors.append(rep_factor)

        return torch.tensor(rep_factors, dtype=torch.float32)

    def _get_epoch_indices(self, generator):
        """
        Create a list of dataset indices (with repeats) to use for one epoch.

        Args:
            generator (torch.Generator): pseudo random number generator used for
                stochastic rounding.

        Returns:
            torch.Tensor: list of dataset indices to use in one epoch. Each index
                is repeated based on its calculated repeat factor.
        """
        # Since repeat factors are fractional, we use stochastic rounding so
        # that the target repeat factor is achieved in expectation over the
        # course of training
        rands = torch.rand(len(self._frac_part), generator=generator)
        rep_factors = self._int_part + (rands < self._frac_part).float()
        # Construct a list of indices in which we repeat images as specified
        indices = []
        for dataset_index, rep_factor in enumerate(rep_factors):
            indices.extend([dataset_index] * int(rep_factor.item()))
        return torch.tensor(indices, dtype=torch.int64)

    def __iter__(self):
        start = self._rank
        yield from itertools.islice(self._infinite_indices(), start, None, self._world_size)

    def _infinite_indices(self):
        g = torch.Generator()
        g.manual_seed(self._seed)
        while True:
            # Sample indices with repeats determined by stochastic rounding; each
            # "epoch" may have a slightly different size due to the rounding.
            indices = self._get_epoch_indices(g)
            if self._shuffle:
                randperm = torch.randperm(len(indices), generator=g)
                yield from indices[randperm]
            else:
                yield from indices


class InferenceSampler(Sampler):
    """
    Produce indices for inference.
    Inference needs to run on the __exact__ set of samples,
    therefore when the total number of samples is not divisible by the number of workers,
    this sampler produces different number of samples on different workers.
    """

    def __init__(self, size: int):
        """
        Args:
            size (int): the total number of data of the underlying dataset to sample from
        """
        self._size = size
        assert size > 0
        self._rank = comm.get_rank()
        self._world_size = comm.get_world_size()

        shard_size = (self._size - 1) // self._world_size + 1
        begin = shard_size * self._rank
        end = min(shard_size * (self._rank + 1), self._size)
        self._local_indices = range(begin, end)

    def __iter__(self):
        yield from self._local_indices

    def __len__(self):
        return len(self._local_indices)



# pylint: disable=wildcard-import
"""Faster-RCNN Object Detection."""
from __future__ import absolute_import

from .faster_rcnn import *
from .doublehead_rcnn import *
from .predefined_models import *
from .rcnn_target import RCNNTargetGenerator, RCNNTargetSampler
from .data_parallel import ForwardBackwardTask



"""Task utils"""
from .classification import train_classification, validation_classification, test_classification
from .coot import train_coot, validate_coot
from .pose import DirectposePipeline, build_pose_optimizer



"""Utils for deep learning framework related functions"""
import numpy as np

__all__ = ['EarlyStopperOnPlateau', '_suggest_load_context', 'create_dummy_estimator']

def _dummy_constructor(self, *arg, **kwargs):
    raise RuntimeError(self.reason.format(type(self).__name__))

def create_dummy_estimator(name, reason):
    assert isinstance(reason, str)
    DummyEstimator = type(name, (object, ), {
        # constructor
        "__init__": _dummy_constructor,

        # data members
        "reason": reason,
    })
    return DummyEstimator


class EarlyStopperOnPlateau:
    """Early stopping on plateau helper.

    Parameters
    ----------
    patience : int, default is -1
        How many epochs with no improvement after which train will be early stopped.
        Negative patience means infinite petience.
    metric_fn : function, default is None
        The function to apply to metric value if any. For example, you can use
        the `metric_fn` to cast loss to negative values where lower loss is better.
        `min_delta`, `baseline_value` and `max_value` are all based on output of `metric_fn`.
    min_delta : float, default is 1e-4
        Early stopper ignores changes less than `min_delta` for metrics to ignore tiny fluctuates.
    baseline_value : float, default is 0.0
        The baseline metric value to be considered.
    max_value : float, default is 1.0
        Instantly early stop if reaching max value.

    """
    def __init__(self, patience=10, metric_fn=None,
                 min_delta=1e-4, baseline_value=None, max_value=np.Inf):
        self.patience = patience if patience > 0 else np.Inf
        self.metric_fn = metric_fn
        self.min_delta = np.abs(min_delta)
        self.baseline_value = baseline_value
        self.max_value = max_value
        self.reset()

    def reset(self):
        """reset the early stopper"""
        self.last_epoch = 0
        self.wait = 0
        self._should_stop = False
        self._message = ''
        if self.baseline_value is not None:
            self.best = self.baseline_value
        else:
            self.best = -np.Inf

    def update(self, metric_value, epoch=None):
        """Update with end of epoch metric.

        Parameters
        ----------
        metric_value : float
            The end of epoch metric.
        epoch : int, optional
            The real epoch in case the update function is not called in every epoch.

        """
        if _is_real_number(epoch):
            if _is_real_number(self.last_epoch):
                diff_epoch = epoch - self.last_epoch
            else:
                diff_epoch = 1
            self.last_epoch = epoch
        else:
            diff_epoch = 1
        if not _is_real_number(metric_value):
            return
        if self.metric_fn is not None:
            metric_value = self.metric_fn(metric_value)

        if metric_value > self.max_value:
            self._should_stop = True
            self._message = 'EarlyStop given {} vs. max {}'.format(metric_value, self.max_value)
        else:
            if metric_value - self.min_delta > self.best:
                self.best = metric_value
                self.wait = 0
            else:
                self.wait += diff_epoch
                if self.wait >= self.patience:
                    self._should_stop = True
                    self._message = 'EarlyStop after {} epochs: no better than {}'.format(self.patience, self.best)

    def get_early_stop_advice(self):
        """Get the early stop advice.

        Returns
        -------
        (bool, str)
            should_stop : bool
                Whether the stopper suggest OnPlateau pattern is active.
            message : str
                The detailed message why early stop is suggested, if `should_stop` is True.
        """
        return self._should_stop, self._message

def _is_real_number(x):
    """Check if x is a real number"""
    return isinstance(x, (int, float, complex)) and not isinstance(x, bool)

def _suggest_load_context(model, mode, orig_ctx):
    """Get the correct context given the mode"""
    if not isinstance(orig_ctx, (list, tuple)):
        orig_ctx = [orig_ctx]
    try:
        import mxnet as mx
    except ImportError:
        mx = None
    try:
        import torch
    except ImportError:
        torch = None
    if mx is not None and isinstance(model, mx.gluon.Block):
        if mode == 'auto':
            if orig_ctx[0].device_type == 'gpu':
                mode = 'gpu'
            else:
                mode = 'cpu'
        if mode == 'cpu':
            return [mx.cpu()]
        if mode == 'gpu':
            return [mx.gpu(i) for i in range(mx.context.num_gpus())]
        if isinstance(mode, (list, tuple)):
            if not all(isinstance(i, int) for i in mode):
                raise ValueError('Requires integer gpu id, given {}'.format(mode))
            return [mx.gpu(i) for i in mode if i in range(mx.context.num_gpus())]
    if torch is not None and isinstance(model, (torch.nn.Module, torch.nn.DataParallel)):
        if mode == 'auto':
            if orig_ctx[0] == torch.device('cpu'):
                mode = 'cpu'
            else:
                mode = 'gpu'
        if mode == 'cpu':
            return [torch.device('cpu')]
        if mode == 'gpu':
            return [torch.device(f'cuda:{gid}') for gid in range(torch.cuda.device_count())]
        if isinstance(mode, (list, tuple)):
            if not all(isinstance(i, int) for i in mode):
                raise ValueError('Requires integer gpu id, given {}'.format(mode))
            return [torch.device(f'cuda:{gid}') for gid in mode if gid in range(torch.cuda.device_count())]

    return None



# adapted from https://github.com/facebookresearch/detectron2/blob/master/detectron2/structures/image_list.py
from __future__ import division
from typing import Any, List, Sequence, Tuple, Union
import torch
from torch.nn import functional as F


class ImageList(object):
    """
    Structure that holds a list of images (of possibly
    varying sizes) as a single tensor.
    This works by padding the images to the same size,
    and storing in a field the original sizes of each image

    Attributes:
        image_sizes (list[tuple[int, int]]): each tuple is (h, w)
    """

    def __init__(self, tensor: torch.Tensor, image_sizes: List[Tuple[int, int]]):
        """
        Arguments:
            tensor (Tensor): of shape (N, H, W) or (N, C_1, ..., C_K, H, W) where K >= 1
            image_sizes (list[tuple[int, int]]): Each tuple is (h, w). It can
                be smaller than (H, W) due to padding.
        """
        self.tensor = tensor
        self.image_sizes = image_sizes

    def __len__(self) -> int:
        return len(self.image_sizes)

    def __getitem__(self, idx: Union[int, slice]) -> torch.Tensor:
        """
        Access the individual image in its original size.

        Returns:
            Tensor: an image of shape (H, W) or (C_1, ..., C_K, H, W) where K >= 1
        """
        size = self.image_sizes[idx]
        return self.tensor[idx, ..., : size[0], : size[1]]  # type: ignore

    def to(self, *args: Any, **kwargs: Any) -> "ImageList":
        cast_tensor = self.tensor.to(*args, **kwargs)
        return ImageList(cast_tensor, self.image_sizes)

    @property
    def device(self) -> torch.device:
        return self.tensor.device

    @staticmethod
    def from_tensors(tensors: Sequence[torch.Tensor], size_divisibility: int = 0, pad_value: float = 0.0):
        """
        Args:
            tensors: a tuple or list of `torch.Tensors`, each of shape (Hi, Wi) or
                (C_1, ..., C_K, Hi, Wi) where K >= 1. The Tensors will be padded
                to the same shape with `pad_value`.
            size_divisibility (int): If `size_divisibility > 0`, add padding to ensure
                the common height and width is divisible by `size_divisibility`.
                This depends on the model and many models need a divisibility of 32.
            pad_value (float): value to pad

        Returns:
            an `ImageList`.
        """
        assert len(tensors) > 0
        assert isinstance(tensors, (tuple, list))
        for t in tensors:
            assert isinstance(t, torch.Tensor), type(t)
            assert t.shape[1:-2] == tensors[0].shape[1:-2], t.shape
        # per dimension maximum (H, W) or (C_1, ..., C_K, H, W) where K >= 1 among all tensors
        max_size = (
            # In tracing mode, x.shape[i] is Tensor, and should not be converted
            # to int: this will cause the traced graph to have hard-coded shapes.
            # Instead we should make max_size a Tensor that depends on these tensors.
            # Using torch.stack twice seems to be the best way to convert
            # list[list[ScalarTensor]] to a Tensor
            torch.stack(
                [
                    torch.stack([torch.as_tensor(dim) for dim in size])
                    for size in [tuple(img.shape) for img in tensors]
                ]
            )
            .max(0)
            .values
        )

        if size_divisibility > 1:
            stride = size_divisibility
            # the last two dims are H,W, both subject to divisibility requirement
            max_size = torch.cat([max_size[:-2], (max_size[-2:] + (stride - 1)) // stride * stride])

        image_sizes = [tuple(im.shape[-2:]) for im in tensors]

        if len(tensors) == 1:
            # This seems slightly (2%) faster.
            # TODO: check whether it's faster for multiple images as well
            image_size = image_sizes[0]
            padding_size = [0, max_size[-1] - image_size[1], 0, max_size[-2] - image_size[0]]
            if all(x == 0 for x in padding_size):  # https://github.com/pytorch/pytorch/issues/31734
                batched_imgs = tensors[0].unsqueeze(0)
            else:
                padded = F.pad(tensors[0], padding_size, value=pad_value)
                batched_imgs = padded.unsqueeze_(0)
        else:
            # max_size can be a tensor in tracing mode, therefore use tuple()
            batch_shape = (len(tensors),) + tuple(max_size)
            batched_imgs = tensors[0].new_full(batch_shape, pad_value)
            for img, pad_img in zip(tensors, batched_imgs):
                pad_img[..., : img.shape[-2], : img.shape[-1]].copy_(img)

        return ImageList(batched_imgs.contiguous(), image_sizes)



from .distributed_sampler import InferenceSampler, RepeatFactorTrainingSampler, TrainingSampler
from .grouped_batch_sampler import GroupedBatchSampler

__all__ = [
    "GroupedBatchSampler",
    "TrainingSampler",
    "InferenceSampler",
    "RepeatFactorTrainingSampler",
]



import tvm
from tvm import relay
from tvm.contrib import graph_executor

from PIL import Image
import numpy as np
import torch
import torchvision.transforms as T
import torchvision
from torchvision import transforms
from gluoncv.torch.engine.config import get_cfg_defaults
from gluoncv.torch import model_zoo

def _test_tvm_export(model, img, target="llvm"):
  model = model.eval()

  # We grab the TorchScripted model via tracing
  input_shape = [1, 3, 224, 224]
  input_data = torch.randn(input_shape)
  scripted_model = torch.jit.trace(model, input_data).eval()
  input_name = "input0"
  shape_list = [(input_name, img.shape)]
  mod, params = relay.frontend.from_pytorch(scripted_model, shape_list)
  if target == "llvm":
    ctx = tvm.cpu(0)
  elif target == "cuda":
    ctx = tvm.gpu(0)
  else:
    raise ValueError(target)
  target_host = "llvm"

  with tvm.transform.PassContext(opt_level=3):
      lib = relay.build(mod, target=target, target_host=target_host, params=params)

  dtype = "float32"
  m = graph_executor.GraphModule(lib["default"](ctx))
  # Set inputs
  m.set_input(input_name, tvm.nd.array(img.astype(dtype)))
  # Execute
  m.run()
  # Get outputs
  tvm_output = m.get_output(0)

# def test_model_zoo_tvm_export():
#   from tvm.contrib.download import download_testdata
#   img_url = "https://github.com/dmlc/mxnet.js/blob/main/data/cat.png?raw=true"
#   img_path = download_testdata(img_url, "cat.png", module="data")
#   img = Image.open(img_path).resize((224, 224))
#   my_preprocess = transforms.Compose(
#     [
#         transforms.Resize(256),
#         transforms.CenterCrop(224),
#         transforms.ToTensor(),
#         transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
#     ]
#   )
#   img = my_preprocess(img)
#   img = np.expand_dims(img, 0)
#
#   model = getattr(torchvision.models, "resnet18")(pretrained=True)
#   _test_tvm_export(model, img)

def test_model_zoo_tvm_export_directpose():
    from gluoncv.torch.utils.tvm_utils.nms import nms
    from tvm.contrib.download import download_testdata
    def get_image(img_name='street_small.jpg', img_url=None, img_width=1280, img_height=736):
        def get_single_image_input(img_url):
            img_name = img_url.split("/")[-1]
            img_path = download_testdata(img_url, img_name, module="data")
            orig_img = Image.open(img_path)
            orig_img = orig_img.resize((img_width, img_height), Image.LANCZOS)
            img = np.array(orig_img)[:, :, (0, 1, 2)].astype('uint8')
            return img, orig_img, img_path

        def get_transforms():
            tforms = T.Compose([T.ToTensor(), T.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])])
            return tforms

        if img_url is None:
            img_url = f"https://raw.githubusercontent.com/dmlc/web-data/master/gluoncv/detection/{img_name}"
        img, orig_img, img_path = get_single_image_input(img_url)

        tforms = get_transforms()
        input_data = tforms(img).unsqueeze(0)
        return input_data, orig_img

    images, orig_img = get_image()
    model_list = ['directpose_resnet50_lpf_fpn_coco']

    cfg = get_cfg_defaults(name='directpose')
    torch.set_grad_enabled(False)
    use_cuda = True
    if torch.cuda.device_count() == 0:
        use_cuda = False
    if use_cuda:
        device = torch.device('cuda')
        target = "cuda"
        ctx = tvm.gpu(0)
    else:
        device = torch.device('cpu')
        target = "llvm"
        ctx = tvm.cpu()
    for model in model_list:
        cfg.CONFIG.MODEL.PRETRAINED = True
        cfg.CONFIG.MODEL.NAME = model
        cfg.CONFIG.MODEL.TVM_MODE = True
        net = model_zoo.get_model(cfg).to(device).eval()

        # y = net(images.to(device))
        with torch.no_grad():
            scripted_model = torch.jit.trace(net.forward, images.to(device)).eval()

        input_name = "input0"
        shape_list = [(input_name, images.shape)]
        mod, params = relay.frontend.from_pytorch(scripted_model, shape_list, {"torchvision::nms": nms})
        target_host = "llvm"
        with tvm.transform.PassContext(opt_level=3):
            lib = relay.build(mod, target=target, target_host=target_host, params=params)

if __name__ == '__main__':
    import nose

    nose.runmodule()
